# coding: utf-8
"""
Show rings in Yaplot format, defined in https://github.com/vitroid/Yaplot

Usage:
    genice III -f rings > 3.rings.yap
    genice III -f rings[5] > output.yap    # up to 5-membered rings.
"""

import sys
from collections import defaultdict
import numpy as np
import networkx as nx
import yaplotlib as yp

from genice import rigid
from countrings import countrings_nx as cr



def face(center, rpos):
    pos = rpos + center
    n = rpos.shape[0]
    s = yp.Color(n)
    s += yp.Layer(n)
    s += yp.Polygon(pos)
    return s



def hook2(lattice):
    lattice.logger.info("Hook2: Show rings in Yaplot format.")
    # copied from svg_poly
    graph = nx.Graph(lattice.graph) #undirected
    cellmat = lattice.repcell.mat
    s = ""
    s += yp.Layer(2)
    s += yp.Color(0)
    for i,j in graph.edges():
        pi, pj = lattice.reppositions[i], lattice.reppositions[j]
        d = pj - pi
        d -= np.floor(d+0.5)
        s += yp.Line(pi @ cellmat, (pi+d) @ cellmat)
    for ring in cr.CountRings(graph).rings_iter(lattice.largestring):
        deltas = np.zeros((len(ring),3))
        d2 = np.zeros(3)
        for k,i in enumerate(ring):
            d = lattice.reppositions[i] - lattice.reppositions[ring[0]]
            d -= np.floor(d+0.5)
            deltas[k] = d
            dd = lattice.reppositions[ring[k]] - lattice.reppositions[ring[k-1]]
            dd -= np.floor(dd+0.5)
            d2 += dd
        # d2 must be zeros
        if np.all(np.absolute(d2) < 1e-5):
            comofs = np.sum(deltas, axis=0) / len(ring)
            deltas -= comofs
            com = lattice.reppositions[ring[0]] + comofs
            com -= np.floor(com)
            # rel to abs
            com    = np.dot(com,    cellmat)
            deltas = np.dot(deltas, cellmat)
            s += face(com,deltas)
    print(s)
    lattice.logger.info("Hook2: end.")


    
# argparser
def hook0(lattice, arg):
    lattice.logger.info("Hook0: ArgParser.")

    if arg == "":
        lattice.largestring=8
    else:
        try:
            lattice.largestring=int(arg)
        except:
            lattice.logger.error("Argument must be a positive integer.")
            sys.exit(1)

    lattice.logger.info("  Largest ring: {0}.".format(lattice.largestring))
    lattice.logger.info("Hook0: end.")



hooks = {2:hook2, 0:hook0}



# coding: utf-8
"""
extended XMol file format (.exyz)

http://open-babel.readthedocs.io/en/latest/FileFormats/Extended_XYZ_cartesian_coordinates_format.html
"""

import numpy as np

from genice import rigid

def hook7(lattice):
    lattice.logger.info("Hook7: Output in extended XYZ format.")
    s = ""
    s += "{0}\n".format(len(lattice.atoms))
    s += "%PBC\n"
    for atom in lattice.atoms:
        molorder, resname, atomname, position, order = atom
        s += "{0:>4}{1:15.5f}{2:15.5f}{3:15.5f}\n".format(atomname,position[0]*10,position[1]*10,position[2]*10)
    s = '#' + "\n#".join(lattice.doc) + "\n" + s
    s += "\n"
    s += "Vector1 {0:15.5f}{1:15.5f}{2:15.5f}\n".format(*lattice.repcell.mat[0]*10)
    s += "Vector2 {0:15.5f}{1:15.5f}{2:15.5f}\n".format(*lattice.repcell.mat[1]*10)
    s += "Vector3 {0:15.5f}{1:15.5f}{2:15.5f}\n".format(*lattice.repcell.mat[2]*10)
    s += "Offset  {0:15.5f}{1:15.5f}{2:15.5f}\n".format(0.,0.,0.)
    print(s,end="")
    lattice.logger.info("Hook7: end.")


hooks = {7:hook7}



import logging
import re

import numpy as np
import pairlist as pl

from genice.cell import rel_wrap


def readaline(file):
    """
    read a non-comment line.
    """
    while True:
        line = file.readline()
        if len(line) == 0:
            return line
        if line[0] != "#":
            return line


def load_iter(file, oname="O", hname=None):
    logger = logging.getLogger()
    while True:
        line = readaline(file)
        if len(line) == 0:
            return
        line = readaline(file)
        if len(line) == 0:
            return
        natom = int(line)
        hatoms = []
        oatoms = []
        skipped = set()
        for i in range(natom):
            line = file.readline()
            # resid = int(line[0:5])
            # resna = line[5:10]
            atomname = line[10:15].replace(' ', '')
            # atomid = int(line[15:20])
            pos = np.array([float(x) for x in line[20:].split()[:3]])  # drop velocity
            if re.fullmatch(oname, atomname):
                oatoms.append(pos)
            elif hname is not None and re.fullmatch(hname, atomname):
                hatoms.append(pos)
            else:
                if atomname not in skipped:
                    logger.info("Skip {0}".format(atomname))
                    skipped.add(atomname)
        c = [float(x) for x in file.readline().split()]
        if len(c) == 3:
            cellmat = np.array([[c[0], 0., 0.],
                                [0., c[1], 0.],
                                [0., 0., c[2]]])
        else:
            cellmat = np.array([[c[0], c[3], c[4]],
                                [c[5], c[1], c[6]],
                                [c[7], c[8], c[2]]])
        celli = np.linalg.inv(cellmat)
        oatoms = np.array(oatoms) @ celli
        if len(hatoms) == 0:
            hatoms = None
        else:
            hatoms = np.array(hatoms) @ celli
        yield oatoms, hatoms, cellmat



# coding: utf-8
import math
import numpy as np


oh = 0.10000
hangle = 109.47 * math.pi / 180 / 2
mass=18
ohz = oh * math.cos(hangle)
ohy = oh * math.sin(hangle)
oz  = -ohz*2/mass


sites = np.array([[0, 0,oz],
                  [0, ohy,ohz+oz],
                  [0,-ohy,ohz+oz]]) # nm

labels = ["Ow","Hw","Hw"]
name = "SOL"



import importlib
import logging
import pkg_resources as pr
import sys
from genice.cell import Cell
import numpy as np
from math import pi, acos

from logging import getLogger, StreamHandler, DEBUG, INFO
logger = getLogger(__name__)
handler = StreamHandler()
handler.setLevel(INFO)
logger.setLevel(INFO)
logger.addHandler(handler)
logger.propagate = False
logging.basicConfig(level=logging.INFO,
                    format="%(levelname)s %(message)s")

name = sys.argv[1]
module = importlib.import_module(name)

cell = Cell(module.cell.mat)
La = np.linalg.norm(cell.mat[0])
Lb = np.linalg.norm(cell.mat[1])
Lc = np.linalg.norm(cell.mat[2])
alpha = acos((cell.mat[1] @ cell.mat[2]) / (Lb * Lc)) * 180 / pi
beta = acos((cell.mat[2] @ cell.mat[0]) / (Lc * La)) * 180 / pi
gamma = acos((cell.mat[0] @ cell.mat[1]) / (La * Lb)) * 180 / pi

s = []
s.append("a={0}".format(La))
s.append("b={0}".format(Lb))
s.append("c={0}".format(Lc))

if alpha != 90.0:
    s.append("A={0}".format(alpha))
    
if beta != 90.0:
    s.append("B={0}".format(beta))
    
if gamma != 90.0:
    s.append("C={0}".format(gamma))

for line in open(name+".py").readlines():
    if line.find("celltype") >= 0:
        pass
    else:
        print(line, end="")

print()
print("from genice.cell import cellvectors")
H = "cell = cellvectors("
print( H + (",\n"+" "*len(H)).join(s) +")")





# coding: utf-8
"""
Test for generating water positions from cage positions
"""


cages="""
12 0.5 0.5 0.5
14 0.5 0.0 -0.25
14 0.0 0.25 0.5
14 0.75 0.5 0.0
14 0.5 0.0 0.25
12 0.0 0.0 0.0
14 0.25 0.5 0.0
14 0.0 -0.25 0.5
"""



from genice         import FrankKasper
from genice.valueparsers import parse_cages
from genice.cell    import cellvectors
cell = cellvectors(a=12.747893943706936,
                   b=12.747893943706936,
                   c=12.747893943706936)
cagepos, cagetype = parse_cages(cages)
waters = [w for w in FrankKasper.toWater(cagepos, cell)]
coord = "relative"
density = FrankKasper.estimate_density(waters, cell, 2.76)
bondlen = 2.76 * 1.2




# coding: utf-8
"""
Test suite: ring direction statistics

Hollins, G. T. Configurational statistics and the dielectric constant of ice. Proc. Phys. Soc. 84, 1001–1016 (1964).
"""

# A directed cycle is expressed as an array of True and False.
# For example,
# --> --> --> : [True, True, False]
# --> <-- --> : [Frue, False, True]
#They are isomorphic if they can be identical by rotations and inversions.

# 


import itertools as it
from countrings import countrings_nx as cr
import sys
from collections import defaultdict
import networkx as nx
from math import log
import fractions
import numpy as np

def isomorphs(a):
    """
    return the symmetry number of cycle a
    """
    iso = set()
    aa = a + a
    for i in range(len(a)):
        part = tuple(aa[i:i+len(a)])
        iso.add(part)
    e = [not x for x in a]
    ee = e + e
    for i in range(len(a)):
        part = tuple(ee[i:i+len(a)])
        iso.add(part)
    return iso


def code(ori):
    """
    from an array of True/False to an integer.
    """
    s = 0
    for i in range(len(ori)):
        if ori[i]:
            s += 1<<i
    return s


def encode(ori):
    m = 9999
    for i in isomorphs(ori):
        c = code(i)
        if c < m:
            m = c
    return m

    
def symmetry(a):
    return len(isomorphs(a))


def freedom(a):
    """
    return possible number of molecular placements
    """
    n = 1
    for i in range(len(a)):
        if a[i-1] is a[i]:
            n *= 2
    return n


def ideal_frequency(a):
    N = len(a)
    return symmetry(a) * freedom(a)


def contains(a,b):
    """
    return True if one of the isomorphs of b is included in the set a.
    """
    return len(a & isomorphs(b)) > 0
    

def ideal_frequencies(N):
    freq = dict()
    for a in it.product((False, True), repeat=N):
        c = encode(a)
        if c not in freq:
            freq[c] = ideal_frequency(a)
    return freq


def probabilities(N):
    freq = ideal_frequencies(N)
    sum = 0
    for v in freq:
        sum += freq[v]
    prob = dict()
    for v in freq:
        prob[v] = fractions.Fraction(freq[v], sum)
    return prob


def orientations(members, digraph):
    return [digraph.has_edge(members[i-1], members[i]) for i in range(len(members))]


def spanning(ring, rpos):
    dsum = np.zeros(3)

    for k in range(len(ring)):
        d = rpos[ring[k-1]] - rpos[ring[k]]
        d -= np.floor(d+0.5)
        dsum += d

    return not np.all(np.absolute(dsum) < 1e-5)
    



def hook4(lattice):
    lattice.logger.info("Hook4: Statistics on the HBs along a ring.")
    graph = nx.Graph(lattice.spacegraph) #undirected
    stat = dict()
    prob = defaultdict(int)

    # Ideal distributions
    for n in range(3, lattice.largestring+1):
        prob[n] = probabilities(n)
        stat[n] = defaultdict(int)

    for ring in cr.CountRings(graph).rings_iter(lattice.largestring):
        if not spanning(ring, lattice.reppositions):
            ori = orientations(ring, lattice.spacegraph)
            c   = encode(ori)
            n   = len(ring)
            stat[n][c] += 1

    #size code code(binary) Approx. Stat.
    lattice.logger.info("""
    _ringstat plugin makes the statistics of bond orientations along each
    HB ring and compare the distribution with that of an ideal (isolated
    random) ring. The difference in the distribution is evaluated by
    Kullback-Leibler # divergence, d_{KL}.
    A typical dKL is zero for hydrogen-disordered ices, while it is
    larger than 1 for hydrogen-ordered ones like ices 2 and 9.

    Ringstat analysis validates the ring-scale randomness. GenIce tool
    also certifies the zero net dipole moment and Bernal-Fowler-Pauling
    ice rule in terms of the validity in global and local structures.

    Columns in the output:
      ring size
      code (decimal) indicating the orientations of the bonds.
      code (binary)
      expectation (fractional) for an isolated random ring
      expectation (numerical)
      observation (fractional) in the given structure
      observation (numerical)

    dKL between expectiations and observations is also calculated.
    """)
    for n in range(3, lattice.largestring+1):
        fmtstr = "{{0}} {{1}} {{1:0{0}b}} {{2}} {{3:.5f}} {{5}}/{{6}} {{4:.5f}} ".format(n)
        denom = 0
        for c in stat[n]:
            denom += stat[n][c]
        if denom > 0:
            dKL = 0.0
            for c in prob[n]:
                print(fmtstr.format(n,c,prob[n][c], float(prob[n][c]), stat[n][c]/denom, stat[n][c], denom))
                q = stat[n][c]/denom
                p = prob[n][c]
                if q > 0.0:
                    dKL += q*(log(q) - log(p))
            dKL /= log(2.0)  #bit
            print("{1} {0} dKL[{0}-ring]".format(n,dKL))
            lattice.logger.info("  dKL[{0}-ring]={1}".format(n,dKL))
    #print(score)
    lattice.logger.info("Hook4: end.")


# argparser
def hook0(lattice, arg):
    lattice.logger.info("Hook0: ArgParser.")

    if arg == "":
        lattice.largestring=8
    else:
        try:
            lattice.largestring=int(arg)
        except:
            lattice.logger.error("Argument must be a positive integer.")
            sys.exit(1)

    lattice.logger.info("  Largest ring: {0}.".format(lattice.largestring))
    lattice.logger.info("Hook0: end.")


hooks = {4:hook4, 0:hook0}



#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function
from math import *
import sys
import numpy as np
import numpy.linalg
import logging

# Rigid.py uses vertical vectors while other pars of genice uses horizontal vectors.

# v1,v2,v3: numpy arrays
# return value: fourth vector candidate


def decide1(v1, v2, v3):
    v4 = -(v1 + v2 + v3)
    return v4 / np.linalg.norm(v4)


def decide2(v1, v2):
    d = v1 + v2
    d /= np.linalg.norm(d)
    v = np.cross(v1, v2)
    v /= np.linalg.norm(v)
    d *= -sqrt(1. / 3)
    v *= sqrt(2. / 3)
    return d + v, d - v


# outer product of two vectors; return None if vector is too small
def op(i, j, check=True):
    if check and (np.linalg.norm(i) < 0.001 or np.linalg.norm(j) < 0.001):
        return None
    a = np.cross(i, j)
    if check and np.linalg.norm(a) < 0.001:
        return None
    return a


# calculate quaternions from a rotation matrix (three orthogonal unit vectors)
def rotmat2quat0(i, j, k):
    # print sqlen(i),sqlen(j),sqlen(k)
    ex = np.array((1.0, 0.0, 0.0))
    ey = np.array((0.0, 1.0, 0.0))
    ez = np.array((0.0, 0.0, 1.0))

    # i軸をx軸に移す回転の軸は、iとxの2分面上にある。
    # j軸をy軸に移す回転の軸は、jとyの2分面上にある。
    # そして、それらを同時にみたす回転の軸は、2つの2分面の交線である。
    # 交線は、2つの面の法線のいずれとも直交する=外積である。*/

    a = op(i - ex, j - ey)
    if a is None:
        a = op(i - ex, k - ez)
        if a is None:
            a = op(k - ez, j - ey)
            if a is None:
                #sys.stderr.write("outer prod warning\n")
                # //全く回転しないケース
                return 1.0, 0.0, 0.0, 0.0
    a /= np.linalg.norm(a)
    # /*回転軸aが求まったので、x軸をi軸に重ねる回転の大きさを求める。。*/
    x0 = ex - a[0] * a
    i0 = i - a[0] * a
    if np.linalg.norm(i0) < 0.1:
        i0 = j - a[1] * a
        x0 = ey - a[1] * a

    i0 /= np.linalg.norm(i0)
    x0 /= np.linalg.norm(x0)
    cosine = np.dot(i0, x0)
    if cosine < -1.0 or cosine > 1.0:
        cosh = 0.0
        sinh = 1.0
    else:
        cosh = sqrt((1.0 + cosine) * 0.5)
        sinh = sqrt(1.0 - cosh * cosh)
    o = op(i0, x0, False)
    if np.dot(o, a) < 0.0:
        sinh = -sinh

    return np.array((cosh, -sinh * a[0], +sinh * a[1], -sinh * a[2]))


def rotmat2quat(m):
    # print "rotmat2quat is not reliable yet."
    # sys.exit(1)
    n = m.transpose()
    return rotmat2quat0(np.array(n[0]), np.array(n[1]), np.array(n[2]))


def quat2rotmat(q):
    a, b, c, d = q
    sp11 = (a * a + b * b - (c * c + d * d))
    sp12 = -2.0 * (a * d + b * c)
    sp13 = 2.0 * (b * d - a * c)
    sp21 = 2.0 * (a * d - b * c)
    sp22 = a * a + c * c - (b * b + d * d)
    sp23 = -2.0 * (a * b + c * d)
    sp31 = 2.0 * (a * c + b * d)
    sp32 = 2.0 * (a * b - c * d)
    sp33 = a * a + d * d - (b * b + c * c)
    return np.array([[sp11, sp12, sp13], [sp21, sp22, sp23], [sp31, sp32, sp33]]).transpose()


def euler2quat(e):
    ea, eb, ec = e
    a = cos(ea / 2) * cos((ec + eb) / 2)
    b = sin(ea / 2) * cos((ec - eb) / 2)
    c = sin(ea / 2) * sin((ec - eb) / 2)
    d = cos(ea / 2) * sin((ec + eb) / 2)
    return np.array((a, b, c, d))


def euler2rotmat(e):
    return quat2rotmat(euler2quat(e))


def rotmat2euler(e):
    return quat2euler(rotmat2quat(e))


def quat2euler(q):
    e = np.zeros(3)
    if q[0] == 1.0:
        return e
    p = 2 * (q[0]**2 + q[3]**2) - 1
    if p > 1.0:
        p = 1.0
    if p < -1.0:
        p = -1.0
    e[0] = acos(p)
    thh = e[0] / 2.0
    sinthh = sin(thh)
    costhh = cos(thh)
    p = q[0] / costhh
    if p > 1.0:
        p = 1.0
    if p < -1.0:
        p = -1.0
    p = acos(p)
    if sinthh == 0.0:
        s = 1.0
    else:
        s = q[1] / sinthh
    if s > 1.0:
        s = 1.0
    if s < -1.0:
        s = -1.0
    s = acos(s)
    if q[3] < 0.0:
        p = 2 * pi - p
    if q[2] > 0:
        e[2] = p + s
        e[1] = p - s
    else:
        e[2] = p - s
        e[1] = p + s
    e[1:3] %= (2 * pi)
    return e


def qadd(q1, q2):
    a1, b1, c1, d1 = q1
    a2, b2, c2, d2 = q2
    a3 = a1 * a2 - b1 * b2 - c1 * c2 - d1 * d2
    b3 = a1 * b2 + b1 * a2 + c1 * d2 - d1 * c2
    c3 = a1 * c2 + c1 * a2 - b1 * d2 + d1 * b2
    d3 = a1 * d2 + d1 * a2 + b1 * c2 - c1 * b2
    if a3 < 0:
        a3 = -a3
        b3 = -b3
        c3 = -c3
        d3 = -d3
    return np.array([a3, b3, c3, d3])


def qmul(q, x):
    if q[0] >= 1.0:
        return np.array(q)
    phi = acos(q[0])
    sine = sqrt(1.0 - q[0]**2)
    if q[1] < 0:
        phi = -phi
        sine = -sine
    phi *= x
    sine = sin(phi) / sine
    a = np.zeros(4)
    a[0] = cos(phi)
    a[1] = q[1] * sine
    a[2] = q[2] * sine
    a[3] = q[3] * sine
    return a


def test_rotation():
    e = np.array((0.2, 0.3, 0.4))
    print(e)
    print(quat2euler(euler2quat(e)))
    print(euler2quat(e))
    print(rotmat2quat(quat2rotmat(euler2quat(e))))
    q = euler2quat(e)
    q = qmul(q, 0.5)
    print(qadd(q, q))


def six2nine(a, b, c, alpha, beta, gamma):
    # convert from angles to matrix
    x = np.array([1.0, 0.0, 0.0])
    y = np.array([0.0, 1.0, 0.0])
    z = np.array([0.0, 0.0, 1.0])
    alpha *= pi / 180
    beta *= pi / 180
    gamma *= pi / 180
    A = a * x
    eb = x * cos(gamma) + y * sin(gamma)
    B = b * eb
    # ec.x = cos(beta)
    # ec.eb = cos(alpha)
    ecx = cos(beta)
    # ecx*cos(gamma)+ecy*sin(gamma)=c cos(alpha)
    ecy = (cos(alpha) - ecx * cos(gamma)) / sin(gamma)
    ecz = (1 - ecx**2 - ecy**2)**0.5
    ec = np.array([ecx, ecy, ecz])
    C = c * ec
    # checked.
    # LA = np.linalg.norm(A)
    # LB = np.linalg.norm(B)
    # LC = np.linalg.norm(C)
    # print(a,b,c)
    # print(LA,LB,LC)
    # p = acos(np.dot(B,C)/(LB*LC))
    # q = acos(np.dot(C,A)/(LC*LA))
    # r = acos(np.dot(A,B)/(LA*LB))
    # print(alpha, beta, gamma)
    # print(p,q,r)
    return np.vstack([A, B, C])


def test():
    """Testing Docstring"""
    six2nine(22.561, 26.318, 25.270, 102.09, 89.66, 89.45)
    print("test 1: identical projection")
    e1 = np.zeros(3)
    q1 = euler2quat(e1)
    r1 = quat2rotmat(q1)
    print("e id", e1)
    print("q id", q1)
    print("r id", r1)
    print("rI to eI", rotmat2euler(r1))
    print()

    print("test 2a: cancellation via quat")
    e1 = np.array([0.0, 0.0, 0.1])
    e2 = np.array([0.0, 0.0, -0.1])
    print("e1", e1)
    print("e2", e2)
    q1 = euler2quat(e1)
    q2 = euler2quat(e2)
    q12 = qadd(q1, q2)
    print("q12", q12)
    e12 = quat2euler(q12)
    print("e1+e2", e12)
    print()

    print("test 3a: euler addition via quat?")
    e1 = np.array([0.1, 0.0, 0.0])
    e2 = np.array([0.2, 0.0, 0.0])
    print("e1", e1)
    print("e2", e2)
    q1 = euler2quat(e1)
    q2 = euler2quat(e2)
    print("++++++++++++++++++++++")
    print("q1", q1)
    print("q2", q2)
    print("++++++++++++++++++++++")
    q12 = qadd(q1, q2)
    print("++++++++++++++++++++++")
    print("q12", q12)
    print("++++++++++++++++++++++")
    e12 = quat2euler(q12)
    print("e1+e2", e12)
    print()

    print("test 3b: euler addition via quat")
    e1 = np.array([0.0, 0.1, 0.0])
    e2 = np.array([0.0, 0.2, 0.0])
    print("e1", e1)
    print("e2", e2)
    q1 = euler2quat(e1)
    q2 = euler2quat(e2)
    print("q1", q1)
    print("q2", q2)
    q12 = qadd(q1, q2)
    print("q12", q12)
    e12 = quat2euler(q12)
    print("e1+e2", e12)
    print()

    print("test 3c: euler addition via quat")
    e1 = np.array([0.0, 0.0, 0.1])
    e2 = np.array([0.0, 0.0, 0.2])
    print("e1", e1)
    print("e2", e2)
    q1 = euler2quat(e1)
    q2 = euler2quat(e2)
    print("q1", q1)
    print("q2", q2)
    q12 = qadd(q1, q2)
    print("q12", q12)
    e12 = quat2euler(q12)
    print("e1+e2", e12)
    r12 = quat2rotmat(q12)
    print("  r12", r12)
    print()

    print("test 4a: euler addition via rotmat")
    e1 = np.array([0.0, 0.0, 0.1])
    e2 = np.array([0.0, 0.0, 0.2])
    print("e1", e1)
    print("e2", e2)
    r1 = euler2rotmat(e1)
    r2 = euler2rotmat(e2)
    r12 = np.dot(r1, r2)
    print("r12", r12)
    q12 = rotmat2quat(r12)
    print("  q12", q12)
    e12 = quat2euler(q12)
    print("e1+e2", e12)
    print()

    q2 = np.array([0.5, 0.5, 0.5, -0.5])
    r1 = quat2rotmat(q1)
    r1i = r1.transpose()
    q1i = rotmat2quat(r1i)
    print("q1", q1)
    print("q1i", q1i)
    print("q1+q1i", qadd(q1i, q1))
    q21 = qadd(q2, qmul(q1, -1))
    print("q21=q1-q2", q21)
    print("q21+q1", qadd(q21, q1))
    q3 = np.array([0.60876143, -0.45804276, 0.45804276, -0.45804276])
    print(qadd(q3, q1))
    print(qadd(q1, q3))
    test_rotation()
    # first molecule by genice 1h -r 1 1 1 --format q
    q4 = np.array([0.0692, -0.2480, 0.2911, -0.9214])
    e4 = quat2euler(q4)
    print(e4)
    qe4 = euler2quat(e4)
    print(qe4)


# http://blog.lostinmyterminal.com/python/2015/05/12/random-rotation-matrix.html
def rand_rotation_matrix(deflection=1.0, randnums=None):
    """
    Creates a random rotation matrix.

    deflection: the magnitude of the rotation. For 0, no rotation; for 1, competely random
    rotation. Small deflection => small perturbation.
    randnums: 3 random numbers in the range [0, 1]. If `None`, they will be auto-generated.
    """
    # from http://www.realtimerendering.com/resources/GraphicsGems/gemsiii/rand_rotation.c

    if randnums is None:
        randnums = np.random.uniform(size=(3,))

    theta, phi, z = randnums

    theta = theta * 2.0 * deflection * np.pi  # Rotation about the pole (Z).
    phi = phi * 2.0 * np.pi  # For direction of pole deflection.
    z = z * 2.0 * deflection  # For magnitude of pole deflection.

    # Compute a vector V used for distributing points over the sphere
    # via the reflection I - V Transpose(V).  This formulation of V
    # will guarantee that if x[1] and x[2] are uniformly distributed,
    # the reflected points will be uniform on the sphere.  Note that V
    # has length sqrt(2) to eliminate the 2 in the Householder matrix.

    r = np.sqrt(z)
    Vx, Vy, Vz = V = (
        np.sin(phi) * r,
        np.cos(phi) * r,
        np.sqrt(2.0 - z)
    )

    st = np.sin(theta)
    ct = np.cos(theta)

    R = np.array(((ct, st, 0), (-st, ct, 0), (0, 0, 1)))

    # Construct the rotation matrix  ( V Transpose(V) - I ) R.

    M = (np.outer(V, V) - np.eye(3)).dot(R)
    return M


if __name__ == "__main__":
    test()



#!/usr/bin/python
"""
Crystallographic data of ice VIII

Kuhs, W. F., Finney, J. L., Vettier, C. & Bliss, D. V. Structure and hydrogen ordering in ices VI, VII, and VIII by neutron powder diffraction. J. Chem. Phys. 81, 3612–3623 (1998).
"""

atoms="""
O 0   0.25       0.1071(12)
D 0   0.4157(15) 0.1935(8)
"""

# space group: I4_1 /a m d No. 141

symops="""
     x,            y,            z
     -x,          1/2-y,          z
    3/4-y,        1/4+x,        3/4+z
    3/4+y,        3/4-x,        1/4+z
     -x,            y,            z
      x,          1/2-y,          z
    1/4+y,        1/4+x,        3/4+z
    1/4-y,        3/4-x,        1/4+z
 
     -x,           -y,           -z
      x,          1/2+y,         -z
    3/4+y,        1/4-x,        3/4-z
    3/4-y,        3/4+x,        1/4-z
      x,           -y,           -z
     -x,          1/2+y,         -z
    1/4-y,        1/4-x,        3/4-z
    1/4+y,        3/4+x,        1/4-z

""".translate({ord(','):''})

# add +1/2, +1/2, +1/2
lines = ""
for line in symops.split("\n"):
    cols = line.split()
    if len(cols) == 3:
        line = " ".join([x + "+1/2" for x in cols]) + "\n"
    lines += line

symops += lines


a=4.656 / 10.0 #nm
b=a
c=6.775 / 10.0 #nm
A=90
B=90
C=90

from genice.cell import cellvectors
cell  = cellvectors(a,b,c,A,B,C)

# helper routines to make from CIF-like data
from genice import CIF
atomd = CIF.atomdic(atoms)
sops  = CIF.symmetry_operators(symops)
waters, pairs = CIF.waters_and_pairs(cell, atomd, sops, rep=(2,2,2))

# All the pairs are fixed.
fixed = pairs

import numpy as np
density = 18*len(waters)/6.022e23 / (np.linalg.det(cell)*1e-21)

coord = "relative"



from logging import getLogger

def line_replacer(line, d):
    logger = getLogger()
    s = ""
    for tag in d:
        loc = line.find(tag)
        if loc >= 0:
            logger.debug("From {0} by {1}.".format(tag, d[tag]))
            replacement = d[tag].splitlines()
            if len(replacement) == 1:
                s = line.replace(tag, replacement[0])
            else:
                indent = line[:loc]
                for newline in replacement:
                    s += indent + newline + "\n"
            return s
    return line




#!/usr/bin/python

"""
**************************************************************************

BÁEZ, L. A. & CLANCY, P. Phase equilibria in extended simple point charge ice‐water systems. J. Chem. Phys. 103, 9744–9755 (1998).

**************************************************************************
"""

atoms="""
O1 0.2382 0.7981 0.3422
O2 0.5000 0.5000 0.0446
H1 -0.2603 0.9742 0.4613
H2 0.1446 0.8545 0.2013
H3 0.4108 0.6151 0.1577
"""

# P2_1 2_1 2
symops="""
     x            y            z
    1/2+x        1/2-y         -z
    1/2-x        1/2+y         -z
     -x           -y            z
"""

# in nm
a,b,c = 0.7176, 0.4428, 0.5040

from genice.cell import cellvectors
cell  = cellvectors(a,b,c)

# helper routines to make from CIF-like data
from genice import CIF
atomd = CIF.atomdic(atoms)
sops  = CIF.symmetry_operators(symops)
# the unit cell is too small to handle; multiply (2,2,2)
waters, fixed = CIF.waters_and_pairs(cell, atomd, sops, rep=(2,2,2))

# set pairs in this way for hydrogen-ordered ices.
pairs = fixed

import numpy as np
density = 18*len(waters)/6.022e23 / (np.linalg.det(cell)*1e-21)

coord = "relative"



# parallelepiped cell


import numpy as np
import sys
import logging
from genice.valueparsers import put_in_array
from math import pi, sin, cos, sqrt, acos


def rel_wrap(relvec):
    return relvec - np.floor(relvec + 0.5)


def rel_wrapf(relvec):
    return relvec - np.floor(relvec)


def sincos(A):
    cA = cos(A)
    if abs(cA) < 1e-16:
        return 1.0, 0.0
    return sin(A), cA


# moved from GenIce/cif/CIF.py
def cellvectors(a,b,c,A=90,B=90,C=90):
    """
    Generate cell vectors from a,b,c and alpha, beta, gamma.
    """

    logger = logging.getLogger()
    A *= pi/180
    B *= pi/180
    C *= pi/180
    sA, cA = sincos(A)
    sB, cB = sincos(B)
    sC, cC = sincos(C)
    ea = np.array([1.0, 0.0, 0.0])
    eb = np.array([cC, sC, 0])
    # ec.ea = ecx = cos(B)
    # ec.eb = ecx*ebx + ecy*eby = cos(A)
    ecx = cB
    ecy = (cA - ecx*eb[0]) / eb[1]
    ecz = sqrt(1-ecx**2-ecy**2)
    ec = np.array([ecx, ecy, ecz])
    # logger.info((cos(A), np.dot(eb, ec)))
    # logger.info((cos(B), np.dot(ec, ea)))
    # logger.info((cos(C), np.dot(ea, eb)))
    return np.vstack([ea*a, eb*b, ec*c])


class Cell():
    mat = np.zeros(9).reshape(3, 3)
    inv = None

    def __init__(self, desc=None):
        self.parse(desc)

    def abs2rel(self, absvecs):
        return np.dot(absvecs, self.inv)

    def rel2abs(self, relvec):
        return np.dot(relvec, self.mat)

    def abs_wrap(self, absvec):
        return self.rel2abs(rel_wrap(self.abs2rel(absvec)))

    def abs_wrapf(self, absvec):
        return self.rel2abs(rel_wrapf(self.abs2rel(absvec)))

    def volume(self):
        return np.linalg.det(self.mat)

    def scale(self, x):
        self.mat *= x
        self.inv = np.linalg.inv(self.mat)

    def scale2(self, x):
        for d in range(3):
            self.mat[d, :] = self.mat[d, :] * x[d]
        self.inv = np.linalg.inv(self.mat)

    def parse(self, mat):
        logger = logging.getLogger()
        logger.debug(("MAT:",mat))
        self.mat = mat.copy()
        La = np.linalg.norm(self.mat[0])
        Lb = np.linalg.norm(self.mat[1])
        Lc = np.linalg.norm(self.mat[2])
        alpha = acos((self.mat[1] @ self.mat[2]) / (Lb * Lc)) * 180 / pi
        beta = acos((self.mat[2] @ self.mat[0]) / (Lc * La)) * 180 / pi
        gamma = acos((self.mat[0] @ self.mat[1]) / (La * Lb)) * 180 / pi
        logging.info("Cell dimension:")
        logging.info("  a = {0}".format(La))
        logging.info("  b = {0}".format(Lb))
        logging.info("  c = {0}".format(Lc))
        logging.info("  A = {0}".format(alpha))
        logging.info("  B = {0}".format(beta))
        logging.info("  C = {0}".format(gamma))
        self.inv = np.linalg.inv(self.mat)

    def serialize_BOX9(self):
        s = "@BOX9\n"
        for d in range(3):
            s += "{0:.4f} {1:.4f} {2:.4f}\n".format(*self.mat[d] * 10)  # AA
        return s



# coding: utf-8
"""
extended XMol file format (.exyz)

http://libatoms.github.io/QUIP/io.html#extendedxyz
"""

import numpy as np

from genice import rigid

def hook7(lattice):
    lattice.logger.info("Hook7: Output in extended XYZ format.")
    s = ""
    s += "{0}\n".format(len(lattice.atoms))
    s += 'Lattice="{0:.3f} {1:.3f} {2:.3f} {3:.3f} {4:.3f} {5:.3f} {6:.3f} {7:.3f} {8:.3f}"\n'.format(*lattice.repcell.mat.reshape((9,))*10)
    for atom in lattice.atoms:
        molorder, resname, atomname, position, order = atom
        s += "{0:>4}{1:15.5f}{2:15.5f}{3:15.5f}\n".format(atomname,position[0]*10,position[1]*10,position[2]*10)
    s = '#' + "\n#".join(lattice.doc) + "\n" + s
    print(s,end="")
    lattice.logger.info("Hook7: end.")


hooks = {7:hook7}



import logging

import numpy as np
import pairlist as pl

from genice import rigid
from genice.molecules import tip4p


def load_iter(file, **kwargs):
    logger = logging.getLogger()
    logger.info("  Loading NX3A assuming TIP4P water.")
    while True:
        line = file.readline()
        if len(line) == 0:
            return
        if len(line) > 4:
            if line[:5] == "@BOX3":
                logger.info("  @BOX3")
                line = file.readline()
                box = np.array([float(x) for x in line.split()[:3]])
                cellmat = np.diag(box) / 10  # in nm
                celli = np.linalg.inv(cellmat)
            elif line[:5] == "@NX3A":
                line = file.readline()
                nmol = int(line.split()[0])
                oatoms = []
                hatoms = []
                logger.info("  @NX3A")
                for i in range(nmol):
                    line = file.readline()
                    cols = line.split()[:6]
                    pos = np.array([float(x) for x in cols[:3]]) / 10
                    # oatoms.append(pos / 10)  # in nm
                    euler = np.array([float(x) for x in cols[3:6]])
                    rotmat = rigid.euler2rotmat(euler)
                    a = tip4p.sites @ rotmat + pos
                    oatoms.append(a[0])
                    hatoms.append(a[1])
                    hatoms.append(a[2])
                oatoms = np.array(oatoms) @ celli
                hatoms = np.array(hatoms) @ celli
                yield oatoms, hatoms, cellmat



#!/usr/bin/env python3
# -*- python -*-

import os
import sys
import argparse as ap
import logging
from genice.importer import safe_import
from genice import genice, analice, __version__, load
import random
import numpy as np


def getoptions():
    parser = ap.ArgumentParser(description='GenIce is a swiss army knife to generate hydrogen-disordered ice structures. (version {0})'.format(__version__), prog='genice')
    parser.add_argument('--version',
                        '-V',
                        action='version',
                        version='%(prog)s {0}'.format(__version__))
    parser.add_argument('--rep',
                        '-r',
                        nargs=3,
                        type=int,
                        dest='rep',
                        default=[1, 1, 1],
                        help='Repeat the unit cell in x,y, and z directions. [1,1,1]')
    parser.add_argument('--dens',
                        '-d',
                        type=float,
                        dest='dens',
                        default=-1,
                        help='Specify the ice density in g/cm3')
    parser.add_argument('--add_noise',
                        type=float,
                        dest='noise',
                        default=0.,
                        metavar='percent',
                        help='Add a Gauss noise with given width (SD) to the molecular positions of water. The value 1 corresponds to 1 percent of the molecular diameter of water.')
    parser.add_argument('--seed',
                        '-s',
                        type=int,
                        dest='seed',
                        default=1000,
                        help='Random seed [1000]')
    parser.add_argument('--format',
                        '-f',
                        dest='format',
                        default="gromacs",
                        metavar="gmeqdypoc",
                        help='Specify file format [g(romacs)|m(dview)|e(uler)|q(uaternion)|d(igraph)|y(aplot)|p(ython module)|o(penScad)|c(entersofmass)|r(elative com)] [gromacs]')
    parser.add_argument('--water',
                        '-w',
                        dest='water',
                        default="tip3p",
                        metavar="model",
                        help='Specify water model. (tip3p, tip4p, etc.) [tip3p]')
    parser.add_argument('--guest',
                        '-g',
                        nargs=1,
                        dest='guests',
                        metavar="D=empty",
                        action="append",
                        help='Specify guest(s) in the cage type. (D=empty, T=co2*0.5+me*0.3, etc.)')
    parser.add_argument('--Guest',
                        '-G',
                        nargs=1,
                        dest='spot_guests',
                        metavar="13=me",
                        action="append",
                        help='Specify guest in the specific cage. (13=me, 32=co2, etc.)')
    parser.add_argument('--Group',
                        '-H',
                        nargs=1,
                        dest='groups',
                        metavar="13=bu-:0",
                        action="append",
                        help='Specify the group. (-H 13=bu-:0, etc.)')
    parser.add_argument('--anion',
                        '-a',
                        nargs=1,
                        dest='anions',
                        metavar="3=Cl",
                        action="append",
                        help='Specify a monatomic anion that replaces a water molecule. (3=Cl, 39=F, etc.)')
    parser.add_argument('--cation',
                        '-c',
                        nargs=1,
                        dest='cations',
                        metavar="3=Na",
                        action="append",
                        help='Specify a monatomic cation that replaces a water molecule. (3=Na, 39=NH4, etc.)')
    parser.add_argument('--visual',
                        dest='visual',
                        default="",
                        metavar="visual",
                        help='Specify the yaplot file to store the depolarization paths. [""]')
    parser.add_argument('--nodep',
                        action='store_true',
                        dest='nodep',
                        default=False,
                        help='No depolarization.')
    parser.add_argument('--asis',
                        action='store_true',
                        dest='asis',
                        default=False,
                        help='Assumes all given HB pairs to be fixed. No shuffle and no depolarization.')
    parser.add_argument('--debug',
                        '-D',
                        action='store_true',
                        dest='debug',
                        help='Output debugging info.')
    parser.add_argument('--quiet',
                        '-q',
                        action='store_true',
                        dest='quiet',
                        help='Do not output progress messages.')
    parser.add_argument('Type',
                        help='Crystal type (1c,1h,etc. See https://github.com/vitroid/GenIce for available ice structures.)')
    return parser.parse_args()


def getoptions_analice():
    parser = ap.ArgumentParser(description='GenIce is a swiss army knife to generate hydrogen-disordered ice structures. (version {0})'.format(__version__), prog='analice', usage='%(prog)s [options]')
    parser.add_argument('--version',
                        '-V',
                        action='version',
                        version='%(prog)s {0}'.format(__version__))
    parser.add_argument('--format',
                        '-f',
                        dest='format',
                        default="gromacs",
                        metavar="gmeqdypoc",
                        help='Specify file format [g(romacs)|m(dview)|e(uler)|q(uaternion)|d(igraph)|y(aplot)|p(ython module)|o(penScad)|c(entersofmass)|r(elative com)] [gromacs]')
    parser.add_argument('--output',
                        '-o',
                        dest='output',
                        metavar="%04d.gro",
                        help='Output in separate files.')
    parser.add_argument('--water',
                        '-w',
                        dest='water',
                        default="tip3p",
                        metavar="model",
                        help='Replace water model. (tip3p, tip4p, etc.) [tip3p]')
    parser.add_argument('--oxygen',
                        '-O',
                        dest='oatom',
                        metavar="OW",
                        default="O",
                        help='Specify atom name of oxygen in input Gromacs file. ("O")')
    parser.add_argument('--hydrogen',
                        '-H',
                        dest='hatom',
                        metavar="HW[12]",
                        default="H",
                        help='Specify atom name (regexp) of hydrogen in input Gromacs file. ("H")')
    parser.add_argument('--suffix',
                        '-s',
                        dest='suffix',
                        metavar="gro",
                        default=None,
                        help='Specify the file suffix explicitly. ((None)')
    parser.add_argument('--filerange',
                        dest='filerange',
                        metavar="[from:]below[:interval]",
                        default="0:1000000",
                        help='Specify the number range for the input filename. ("0:1000000")')
    parser.add_argument('--framerange',
                        dest='framerange',
                        metavar="[from:]below[:interval]",
                        default="0:1000000",
                        help='Specify the number range for the input frames. ("0:1000000")')
    parser.add_argument('--debug',
                        '-D',
                        action='count',
                        dest='debug',
                        help='Output debugging info.')
    parser.add_argument('--quiet',
                        '-q',
                        action='store_true',
                        dest='quiet',
                        help='Do not output progress messages.')
    parser.add_argument('--add_noise',
                        type=float,
                        dest='noise',
                        default=0.,
                        metavar='percent',
                        help='Add a Gauss noise with given width (SD) to the molecular positions of water. The value 1 corresponds to 1 percent of the molecular diameter of water.')
    parser.add_argument('--avgspan', '-v',
                        type=float,
                        dest='avgspan',
                        default=0,
                        metavar='1',
                        help='Output mean atomic positions of a given time span so as to remove fast librational motions and to make a smooth video. The values 0 and 1 specify no averaging.')
    parser.add_argument('File',
                        help='Input file(s). Analice accepts `.gro` (Gromacs), `.mdv` (mdview), and `.nx3a` (rigid rotors) files as input file formats by default. File type is estimated from the suffix. Files of different types cannot be read at a time. File type can be specified explicitly with -s option.')
    return parser.parse_args()


def main():
    # Module-loading paths
    # 1. Look for the modules in the current working directory
    sys.path.append(".")

    # Parse options
    if sys.argv[0].find("analice") >= 0:
        options = getoptions_analice()
        mode = "analice"
    else:
        options = getoptions()
        mode = "genice"

    # Set verbosity level
    if options.debug:
        logging.basicConfig(level=logging.DEBUG,
                            format="%(asctime)s %(levelname)s %(message)s")
    elif options.quiet:
        logging.basicConfig(level=logging.WARN,
                            format="%(levelname)s %(message)s")
    else:
        # normal
        logging.basicConfig(level=logging.INFO,
                            format="%(levelname)s %(message)s")
    logger = logging.getLogger()
    logger.debug("Debug mode.")

    if mode == "genice":
        logger.debug(options.Type)

        lattice_type = options.Type
        seed = options.seed
        rep = options.rep
        density = options.dens
        asis = options.asis
        anions = dict()
        if options.anions is not None:
            logger.info(options.anions)
            for v in options.anions:
                key, value = v[0].split("=")
                anions[int(key)] = value
        cations = dict()
        if options.cations is not None:
            for v in options.cations:
                key, value = v[0].split("=")
                cations[int(key)] = value
        spot_guests = dict()
        if options.spot_guests is not None:
            for v in options.spot_guests:
                key, value = v[0].split("=")
                spot_guests[int(key)] = value
        groups = dict()
        if options.groups is not None:
            for v in options.groups:
                key, value = v[0].split("=")
                groups[int(key)] = value

        # Set random seeds
        random.seed(seed)
        np.random.seed(seed)

        logger.debug("Lattice: {0}".format(lattice_type))
        assert lattice_type is not None

        # Initialize the Lattice class with arguments which are required for plugins.
        lat = genice.GenIce(safe_import("lattice", lattice_type),
                              density=density,
                              rep=rep,
                              cations=cations,
                              anions=anions,
                              spot_guests=spot_guests,
                              spot_groups=groups,
                              asis=asis,
                              )

        water_type = options.water
        guests = options.guests
        noise = options.noise
        depolarize = not options.nodep
        file_format = options.format

        # Main part of the program is contained in th Formatter object. (See formats/)
        logger.debug("Output file format: {0}".format(file_format))
        formatter = safe_import("format", file_format)

        if options.visual != "":
            record_depolarization_path = open(options.visual, "w")
        else:
            record_depolarization_path = None

        del options  # Dispose for safety.

        lat.generate_ice(water_type=water_type,
                         guests=guests,
                         formatter=formatter,
                         record_depolarization_path=record_depolarization_path,
                         noise=noise,
                         depolarize=depolarize,
                         )
    else:  # analice
        logger.debug(options.File)

        water_type = options.water
        file_format = options.format
        oname = options.oatom
        hname = options.hatom
        filename = options.File
        noise = options.noise
        avgspan = options.avgspan
        filerange = options.filerange
        framerange = options.framerange
        suffix = options.suffix
        if options.output is None:
            output = None
            stdout = None
        else:
            output = options.output
            stdout = sys.stdout

        logger.debug(filerange)
        logger.debug(framerange)
        logger.debug(oname)
        logger.debug(hname)
        logger.debug(suffix)
        logger.info("Output:{0}".format(output))

        del options  # Dispose for safety.

        for i, (oatoms, hatoms, cellmat) in enumerate(load.average(lambda:load.iterate(filename, oname, hname, filerange, framerange, suffix=suffix), span=avgspan)):
            # Main part of the program is contained in th Formatter object. (See formats/)
            logger.debug("Output file format: {0}".format(file_format))
            formatter = safe_import("format", file_format)
            lattice_info = load.make_lattice_info(oatoms, hatoms, cellmat)
            lat = analice.AnalIce(lattice_info)
            if output is not None:
                sys.stdout = open(output % i, "w")
            lat.analyze_ice(water_type=water_type,
                            formatter=formatter,
                            noise=noise,
                            )
        if stdout is not None:
            # recover stdout
            sys.stdout = stdout


if __name__ == "__main__":
    main()



# coding: utf-8
"""
Rigid water molecule 
"""

import numpy as np

from genice import rigid


def hook5(lattice):
    lattice.logger.info("Hook5: Output water molecules as rigid rotors (Quaternion).")
    cellmat = lattice.repcell.mat
    s = ""
    if cellmat[1,0] == 0 and cellmat[2,0] == 0 and cellmat[2,1] == 0:
        s += "@BOX3\n"
        s += "{0} {1} {2}\n".format(cellmat[0,0]*10,cellmat[1,1]*10,cellmat[2,2]*10)
    else:
        s += "@BOX9\n"
        for d in range(3):
            s += "{0} {1} {2}\n".format(cellmat[0,d]*10,cellmat[1,d]*10,cellmat[2,d]*10)
    s += "@NX4A\n"
    s += "{0}\n".format(len(lattice.reppositions))
    for pos,rot in zip(lattice.reppositions, lattice.rotmatrices):
        position = np.dot(pos, cellmat)*10   #in Angstrom
        quat     = rigid.rotmat2quat(rot.transpose())
        s += "{0:9.4f} {1:9.4f} {2:9.4f}  {3:9.4f} {4:9.4f} {5:9.4f} {6:9.4f}\n".format(position[0],
                                                                        position[1],
                                                                        position[2],
                                                                        quat[0],
                                                                        quat[1],
                                                                        quat[2],
                                                                        quat[3])
    s = "\n".join(lattice.doc) + "\n" + s
    print(s,end="")
    lattice.logger.info("Hook5: end.")


hooks = {5:hook5}



#!/usr/bin/python

"""
Data source

BÁEZ, L. A. & CLANCY, P. Phase equilibria in extended simple point charge ice‐water systems. J. Chem. Phys. 103, 9744–9755 (1998).
"""

atoms="""
O1 0.0995 0.1901 0.2835
O2 0.3431 0.3469 0.7598
O3 0.4055 0.0558 0.5013
H1 -0.0149 0.1524 0.3658
H2 0.0921 0.3365 0.2570
H3 0.4441 0.3484 0.8630
H4 0.3713 0.2326 0.6746
H5 0.3045 0.1078 0.4119
H6 0.3847 -0.0897 0.5194
"""

# P4_1, No. 76
symops="""
      x            y            z
     -x           -y          1/2+z
     -y            x          1/4+z
      y           -x          3/4+z
"""

# in nm
a,b,c = 0.6733, 0.6733, 0.7164

from genice.cell import cellvectors
cell  = cellvectors(a,b,c)

# helper routines to make from CIF-like data
from genice import CIF
atomd = CIF.atomdic(atoms)
sops  = CIF.symmetry_operators(symops)
waters, fixed = CIF.waters_and_pairs(cell, atomd, sops)

# set pairs in this way for hydrogen-ordered ices.
pairs = fixed

import numpy as np
density = 18*len(waters)/6.022e23 / (np.linalg.det(cell)*1e-21)

coord = "relative"



from logging import getLogger
import re
from pathlib import Path
from genice.importer import safe_import
from genice.cell import rel_wrap
import numpy as np
from types import SimpleNamespace
import pairlist as pl

def str2rangevalues(s):
    values = s.split(":")
    assert len(values) > 0
    if len(values) == 1:
        return [0, int(values[0]), 1]
    elif len(values) == 2:
        return [int(values[0]), int(values[1]), 1]
    else:
        return [int(v) for v in values]


def str2range(s):
    return range(*str2rangevalues(s))


def iterate(filename, oname, hname, filerange, framerange, suffix=None):
    logger = getLogger()
    rfile = str2range(filerange)
    rframe = str2rangevalues(framerange)
    logger.info("  file number range: {0}:{1}:{2}".format(*str2rangevalues(filerange)))
    logger.info("  frame number range: {0}:{1}:{2}".format(*rframe))
    # test whether filename has a regexp for enumeration
    logger.info(filename)
    m = re.search("%[0-9]*d", filename)
    # prepare file list
    if m is None:
        filelist = [filename, ]
    else:
        filelist = []
        for num in rfile:
            fname = filename % num
            if os.path.exists(fname):
                filelist.append(fname)
    logger.debug("File list: {0}".format(filelist))
    frame = 0
    for fname in filelist:
        logger.info("  File name: {0}".format(fname))
        # single file may contain multiple frames
        if suffix is None:
            suffix = Path(fname).suffix[1:]
        loader = safe_import("loader", suffix)
        file = open(fname)
        for oatoms, hatoms, cellmat in loader.load_iter(file, oname=oname, hname=hname):
            if frame == rframe[0]:
                logger.info("Frame: {0}".format(frame))
                yield oatoms, hatoms, cellmat
                rframe[0] += rframe[2]
                if rframe[1] <= rframe[0]:
                    return
            else:
                logger.info("Skip frame: {0}".format(frame))
            frame += 1


def average(load_iter, span=0):
    logger = getLogger()
    if span <= 1:
        # pass through.
        yield from load_iter()
        return
    ohist = [] # center-of-mass position (relative)
    for oatoms, hatoms, cellmat in load_iter():

        # center of mass
        if len(ohist) == 0:
            # first ohist; just store
            ohist.append(oatoms)
        elif span > 1:
            # displacements
            d = oatoms - ohist[-1]
            d -= np.floor(d+0.5)
            # new positions
            ohist.append(ohist[-1]+d)
            # if too many storage
            if len(ohist) > span:
                # drop the oldest one.
                ohist.pop(0)
            # overwrite the water positions with averaged ones.
            oatoms = np.average(np.array(ohist), axis=0)
        else:
            ohist[0] = oatoms

        yield oatoms, hatoms, cellmat


# def history(load_iter, span=0):
    

def make_lattice_info(oatoms, hatoms, cellmat):
    logger = getLogger()

    assert oatoms.shape[0] > 0
    assert hatoms is None or oatoms.shape[0] * 2 == hatoms.shape[0]
    coord = 'relative'
    density = oatoms.shape[0] / (np.linalg.det(cellmat) * 1e-21) * 18 / 6.022e23

    if hatoms is None:
        return SimpleNamespace(waters=oatoms, coord=coord, density=density, bondlen=0.3, cell=cellmat)

    rotmat = np.zeros((oatoms.shape[0],3,3))
    for i in range(oatoms.shape[0]):
        ro = oatoms[i]
        rdh0 = rel_wrap(hatoms[i * 2] - ro)
        rdh1 = rel_wrap(hatoms[i * 2 + 1] - ro)
        dh0 = np.dot(rdh0, cellmat)
        dh1 = np.dot(rdh1, cellmat)
        y = dh0 - dh1
        y /= np.linalg.norm(y)
        z = dh0 + dh1
        z /= np.linalg.norm(z)
        x = np.cross(y, z)
        rotmat[i] = np.vstack([x, y, z])
        # 重心位置を補正。
        oatoms[i] += (rdh0 + rdh1) * 1. / 18.
    grid = pl.determine_grid(cellmat, 0.245)
    # remove intramolecular OHs
    # 水素結合は原子の平均位置で定義している。
    pairs = []
    logger.debug("  Make pair list.")
    for o, h in pl.pairs_fine_hetero(oatoms, hatoms, 0.245, cellmat, grid, distance=False):
        if not (h == o * 2 or h == o * 2 + 1):
            # hとoは別の分子の上にあって近い。
            # register a new intermolecular pair
            pairs.append((h // 2, o))
    logger.debug("  # of pairs: {0} {1}".format(len(pairs), oatoms.shape[0]))
        
    return SimpleNamespace(waters=oatoms, coord=coord, density=density, pairs=pairs, rotmat=rotmat, cell=cellmat, __doc__=None)

    



import numpy as np
import logging
from collections import Iterable


def flatten(item):
    """Yield items from any nested iterable; see REF."""
    if isinstance(item, str):
        yield item
    elif isinstance(item, Iterable):
        for x in item:
            yield from flatten(x)
    else:
        yield item


def parse_cages(cages):
    cagetype = []
    cagepos = []
    if isinstance(cages, str):
        for line in cages.split("\n"):
            cols = line.split()
            if len(cols) > 0:
                cagetype.append(cols[0])
                cagepos.append([float(x) for x in cols[1:4]])
        cagepos = np.array(cagepos)
    else:
        # Assume it is a list of list
        # flatten
        c = list(flatten(cages))
        while len(c):
            cagetype.append(c.pop(0))
            cagepos.append(c[:3])
            c = c[3:]
        cagepos = np.array(cagepos)
    cagepos -= np.floor(cagepos)
    return cagepos, cagetype


def parse_pairs(values):
    if isinstance(values, str):
        lines = values.split("\n")
        pairs = []
        for line in lines:
            columns = line.split()
            if len(columns) == 2:
                i, j = [int(x) for x in columns]
                pairs.append((i, j))
        return pairs
    elif isinstance(values, list):
        pairs = []
        for pair in values:
            pairs.append(tuple(pair[:2]))  # Make it a tuple
        return pairs


def put_in_array(v):
    """
    Obtain a Numpy array from any kind of values.
    """
    if isinstance(v, str):
        return np.fromstring(v, sep=" ")
    elif isinstance(v, list):
        return np.array(v)
    else:
        return v



#!/usr/bin/python
"""
Helpers for lattice plugin.

Generate a lattice from CIF-like data.
"""


import numpy as np
from math import *
import pairlist as pl
from collections import defaultdict

from logging import getLogger



def fullatoms(atomd, sops):
    global x,y,z
    logger = getLogger()
    # the variables to be evaluated by eval() must be global (?)
    full = []
    names = []
    for name, pos in atomd.items():
        x,y,z = pos
        X,Y,Z = x,y,z # alias
        for sop in sops:
            # print(x,sop)
            p = np.array([eval(s) for s in sop])
            p -= np.floor(p)
            tooclose = False
            for f in full:
                d = f - p
                d -= np.floor(d+0.5)
                L2 = np.dot(d,d)
                if L2 < 0.0001:
                    # print(f,p)
                    logger.debug("Too close: {0} {1}".format(f,p))
                    tooclose = True
                    break
            if not tooclose:
                # print(p, (x,y,z), sop)
                yield name, p
                full.append(p)
                names.append(name)




def atomdic(atoms):
    atomd = dict()
    for atom in atoms.split("\n"):
        cols = atom.split()
        if len(cols) == 0:
            continue
        # remove error digits from 1-3 cols.
        xyz = []
        for col in cols[1:4]:
            p = col.find("(")
            if p >= 0:
                col = col[:p]
            xyz.append(float(col))
        name = cols[0]
        atomd[name] = xyz
    return atomd


def symmetry_operators(symops):
    sops = []
    for symop in symops.split("\n"):
        cols = symop.split()
        if len(cols) == 0:
            continue
        ops = []
        for col in cols:
            if col[-1] == ",":
                col = col[:-1]
            ops.append(col)
        sops.append(ops)
    return sops


def waters_and_pairs(cell, atomd, sops, rep=(1,1,1)):

    logger = getLogger()

    oxygens = []
    hydrogens = []
    for name, pos in fullatoms(atomd, sops):
        if name[0] in ("O",):
            oxygens.append(pos)
        elif name[0] in "DH":
            hydrogens.append(pos)

    assert len(oxygens)*2 == len(hydrogens) or len(hydrogens) == 0, "{0}:{1}".format(len(oxygens)*2,len(hydrogens))

    cell *= np.array(rep)
    
    oo = [[o[0]+x, o[1]+y, o[2]+z]
           for o in oxygens
           for x in range(rep[0])
           for y in range(rep[1])
           for z in range(rep[2])]
    oxygens = np.array(oo)
    oxygens /= np.array(rep)

    if len(hydrogens) == 0:
        return oxygens, None

    hh = [[h[0]+x, h[1]+y, h[2]+z]
           for h in hydrogens
           for x in range(rep[0])
           for y in range(rep[1])
           for z in range(rep[2])]
    hydrogens = np.array(hh)
    hydrogens /= np.array(rep)

    oh = defaultdict(list)
    parent = dict()
    grid = pl.determine_grid(cell, 0.15)
    # find covalent OH bonds
    for i,j in pl.pairs_fine_hetero(oxygens, hydrogens, 0.15, cell, grid, distance=False):
        oh[i].append(j)
        parent[j] = i
    
    grid = pl.determine_grid(cell, 0.20)
    pairs = []
    # find HBs
    for i,j in pl.pairs_fine_hetero(oxygens, hydrogens, 0.20, cell, grid, distance=False):
        if j not in oh[i]:
            # H is on a different water molecule
            p = parent[j]
            pairs.append([p, i])

    logger.debug(pairs)

    waters = []
    for i in range(len(oh)):
        logger.debug((i, oh[i]))
        j,k = oh[i]
        dhj = hydrogens[j] - oxygens[i]
        dhk = hydrogens[k] - oxygens[i]
        dhj -= np.floor(dhj+0.5)
        dhk -= np.floor(dhk+0.5)
        com = oxygens[i] + (dhj + dhk)/18
        waters.append(com)

    return waters, pairs



import sys
import logging
import random
import itertools as it
import logging
import os
from math import sin, cos, pi
from collections import defaultdict

import numpy as np
import pairlist as pl

from genice.importer import safe_import
from genice import digraph as dg
from genice import rigid
from genice.cell import rel_wrap, Cell
from genice.valueparsers import parse_cages, parse_pairs, put_in_array, flatten

# for alkyl groups (Experimental)
from genice import alkyl


def assume_tetrahedral_vectors(v):
    """
    Assume missing vectors at a tetrahedral node.

    Given: known vectors.
    Returns: assumed vectors
    """

    assert len(v) > 0

    if len(v) == 3:

        return [-(v[0] + v[1] + v[2])]

    elif len(v) == 2:
        y = v[1] - v[0]
        y /= np.linalg.norm(y)
        z = v[1] + v[0]
        z /= np.linalg.norm(z)
        x = np.cross(y, z)
        v2 = (x * 8.0**0.5 - z) / 3.0
        v3 = (-x * 8.0**0.5 - z) / 3.0

        return [v2, v3]

    elif len(v) == 1:
        vr = np.array([random.random() for i in range(3)])
        vr /= np.linalg.norm(vr)
        z = v[0] / np.linalg.norm(v[0])
        x = np.cross(z, vr)
        y = np.cross(z, x)
        x1 = -x / 2 + 3.0**0.5 * y / 2
        x2 = -x / 2 - 3.0**0.5 * y / 2

        return [x, x1, x2]

    return []


def orientations(coord, graph, cell):
    """
    Does not work when two OHs are colinear
    """

    logger = logging.getLogger()
    rotmatrices = []
    assert len(coord) == graph.number_of_nodes()  # just for a test of pure water

    for node in range(graph.number_of_nodes()):
        if node in graph.ignores:
            # for dopants; do not rotate
            rotmat = np.identity(3)
        else:
            vsucc = [cell.rel2abs(rel_wrap(coord[x] - coord[node])) for x in graph.successors(node)]

            if len(vsucc) < 2:  # TSL
                vpred = [cell.rel2abs(rel_wrap(coord[x] - coord[node])) for x in graph.predecessors(node)]
                vsucc = [x / np.linalg.norm(x) for x in vsucc]
                vpred = [x / np.linalg.norm(x) for x in vpred]

                if len(vpred) > 2:
                    vpred = vpred[:2]  # number of incoming bonds should be <= 2
                vcomp = assume_tetrahedral_vectors(vpred + vsucc)
                logger.debug("Node {0} vcomp {1} vsucc {2} vpred {3}".format(node, vcomp, vsucc, vpred))
                vsucc = (vsucc + vcomp)[:2]

            logger.debug("Node {0} vsucc {1}".format(node, vsucc))
            assert 2 <= len(vsucc), "Probably a wrong ice network."
            y = vsucc[1] - vsucc[0]
            y /= np.linalg.norm(y)
            z = (vsucc[0] + vsucc[1]) / 2
            z /= np.linalg.norm(z)
            x = np.cross(y, z)
            rotmat = np.vstack([x, y, z])

        rotmatrices.append(rotmat)

    return rotmatrices


def arrange_atoms(coord, cell, rotmatrices, intra, labels, name, ignores=set()):
    atoms = []

    if len(intra) == 0:
        return atoms

    for order, pos in enumerate(coord):
        if order in ignores:
            continue

        abscom = cell.rel2abs(pos)  # relative to absolute
        rotated = np.dot(intra, rotmatrices[order])

        for i in range(len(labels)):
            atoms.append([i, name, labels[i], rotated[i, :] + abscom, order])

    return atoms


def shortest_distance(coord, cell, pairs=None):
    dmin = 1e99

    if pairs is None:
        iter = it.combinations(coord, 2)
    else:
        iter = [(coord[i], coord[j]) for i, j in pairs]

    for c1, c2 in iter:
        r = cell.rel2abs(rel_wrap(c1 - c2))
        rr = np.dot(r, r)

        if rr < dmin:
            dmin = rr

    return dmin**0.5


def replicate_groups(groups, waters, cagepos, rep):
    """
    This is not that easy.
    """
    logger = logging.getLogger()
    # Storage for replicated groups
    newgroups = defaultdict(dict)

    for root, cages in groups.items():
        # Position of root (water) (fractional)
        root_pos = waters[root]

        for cage, group_name in cages.items():
            # Position of the cage (fractional)
            cage_pos = cagepos[cage]
            # Relative position of the cage
            delta = rel_wrap(cage_pos - root_pos)
            # (Image) cell that the cage resides
            gcell = np.floor(root_pos + delta)

            for x in range(rep[0]):
                for y in range(rep[1]):
                    for z in range(rep[2]):
                        r = np.array((x, y, z))
                        # label of the root (water) in the replica
                        newroot = root + len(waters) * \
                            (x + rep[0] * (y + rep[1] * z))
                        # replicated cell in which the cage resides.
                        # modulo by positive number is always positive.
                        cr = (r + gcell) % rep
                        newcage = cage + \
                            len(cagepos) * (cr[0] + rep[0]
                                            * (cr[1] + rep[1] * cr[2]))
                        newcage = int(newcage)
                        newgroups[newroot][newcage] = group_name
                        # logger.info(("root",newroot,"newcage", newcage))
    return newgroups


def replicate_labeldict(labels, nmol, rep):
    newlabels = dict()

    for j in labels:
        for x in range(rep[0]):
            for y in range(rep[1]):
                for z in range(rep[2]):
                    newj = j + nmol * (x + rep[0] * (y + rep[1] * z))
                    newlabels[newj] = labels[j]

    return newlabels


def replicate_labels(labels, nmol, rep):
    newlabels = set()

    for j in labels:
        for x in range(rep[0]):
            for y in range(rep[1]):
                for z in range(rep[2]):
                    newj = j + nmol * (x + rep[0] * (y + rep[1] * z))
                    newlabels.add(newj)

    return newlabels


def replicate_graph(graph, positions, rep):
    repgraph = dg.IceGraph()
    nmol = positions.shape[0]

    for i, j in graph.edges(data=False):
        # positions in the unreplicated cell
        vec = positions[j] - positions[i]
        delta = np.floor(vec + 0.5).astype(int)

        for x in range(rep[0]):
            for y in range(rep[1]):
                for z in range(rep[2]):
                    xi = (x + delta[0] + rep[0]) % rep[0]
                    yi = (y + delta[1] + rep[1]) % rep[1]
                    zi = (z + delta[2] + rep[2]) % rep[2]
                    newi = i + nmol * (xi + rep[0] * (yi + rep[1] * zi))
                    newj = j + nmol * (x + rep[0] * (y + rep[1] * z))

                    if graph[i][j]['fixed']:  # fixed pair
                        repgraph.add_edge(newi, newj, fixed=True)
                    else:

                        # shuffle the bond directions
                        if 0 == random.randint(0, 1):
                            repgraph.add_edge(newi, newj, fixed=False)
                        else:
                            repgraph.add_edge(newj, newi, fixed=False)

    # replicate "ignores" == dopants list in the graph
    repgraph.ignores = replicate_labels(graph.ignores, nmol, rep)

    return repgraph


def replicate_positions(positions, rep):
    repx = positions.copy()

    for m in range(1, rep[0]):
        v = np.array([m, 0, 0])
        repx = np.concatenate((repx, positions + v))

    repy = repx.copy()

    for m in range(1, rep[1]):
        v = np.array([0, m, 0])
        repy = np.concatenate((repy, repx + v))

    repz = repy.copy()

    for m in range(1, rep[2]):
        v = np.array([0, 0, m])
        repz = np.concatenate((repz, repy + v))

    return repz / rep


def neighbor_cages_of_dopants(dopants, waters, cagepos, cell):
    """
    Just shows the environments of the dopants
    """
    #logger = logging.getLogger()
    dnei = defaultdict(set)

    for site, name in dopants.items():
        org = waters[site]

        for i, pos in enumerate(cagepos):
            #Displacement (relative)
            a = cell.rel2abs(rel_wrap(pos - org))
            sqdistance = np.dot(a, a)

            if sqdistance < 0.57**2:
                dnei[site].add(i)
                # logger.info((i,cagepos[i]))

    return dnei


# They should be separate plugins in the future.
def butyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", "Md"]]])


def pentyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", ["Md", "Me"]]]])


def propyl(cpos, root, cell, molname):
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", "Mc"]])


def ethyl(cpos, root, cell, molname):
    return Alkyl(cpos, root, cell, molname, ["Ma", "Mb"])


def _2_2_dimethylpropyl(cpos, root, cell, molname):
    """
    2,2-dimethylpropyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", "Mc", "Md", "Me"]])


def _2_3_dimethylbutyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", "Md", "Me"], "Mf"]])


def _3_methylbutyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", "Md", "Me"]]])


def _3_3_dimethylbutyl(cpos, root, cell, molname):
    """
    put a butyl group rooted at root toward cpos.
    """
    return Alkyl(cpos, root, cell, molname, ["Ma", ["Mb", ["Mc", "Md", "Me", "Mf"]]])


def Alkyl(cpos, root, cell, molname, backbone):
    """
    put a normal-alkyl group rooted at root toward cpos.
    """
    logger = logging.getLogger()
    # logger.info("  Put butyl at {0}".format(molname))
    v1abs = cell.rel2abs(rel_wrap(cpos - root))
    v1 = v1abs / np.linalg.norm(v1abs)

    origin = cell.rel2abs(root)
    CC = 0.154
    rawatoms = alkyl.alkyl(v1, v1abs * 1.5 / CC, backbone)

    atoms = []
    for i, atom in enumerate(rawatoms):
        atomname, pos = atom
        atompos = cell.abs_wrapf(pos * CC + origin)
        atoms.append([i, molname, atomname, atompos, 0])

    return atoms


class GenIce():
    def __init__(self,
                 lat,
                 density=0,
                 rep=(1, 1, 1),
                 cations=dict(),
                 anions=dict(),
                 spot_guests=dict(),
                 spot_groups=dict(),
                 asis=False,
                 ):

        self.logger = logging.getLogger()
        self.rep = rep
        self.asis = asis
        self.cations = cations
        self.anions = anions
        self.spot_guests = spot_guests
        self.spot_groups = spot_groups
        # Show the document of the module

        try:
            self.doc = lat.__doc__.splitlines()
        except BaseException:
            self.doc = []

        self.doc.append("")
        self.doc.append("Command line: {0}".format(" ".join(sys.argv)))

        for line in self.doc:
            self.logger.info("  "+line)
        # ================================================================
        # rotmatrices (analice)
        #
        try:
            self.rotmatrices = lat.rotmat
        except BaseException:
            self.logger.info("No rotmatrices in lattice")
            self.rotmatrices = None
        # ================================================================
        # waters: positions of water molecules
        #
        self.waters = put_in_array(lat.waters)
        self.logger.debug("Waters: {0}".format(len(self.waters)))
        self.waters = self.waters.reshape((self.waters.size // 3, 3))

        # ================================================================
        # cell: cell dimension
        #   see parse_cell for syntax.
        #
        self.cell = Cell(lat.cell)

        # ================================================================
        # coord: "relative" or "absolute"
        #   Inside genice, molecular positions are always treated as "relative"
        #
        if lat.coord == "absolute":
            self.waters = self.cell.abs2rel(self.waters)

        self.waters = np.array([w - np.floor(w) for w in self.waters])

        # ================================================================
        # pairs: specify the pairs of molecules that are connected.
        #   Bond orientation will be shuffled later
        #   unless it is "fixed".
        #
        self.pairs = None

        try:
            self.pairs = parse_pairs(lat.pairs)
        except AttributeError:
            self.logger.info("HB connectivity is not defined.")

        # ================================================================
        # bondlen: specify the bond length threshold.
        #   This is used when "pairs" are not specified.
        #   It is applied to the original positions of molecules (before density setting).
        #
        nmol = self.waters.shape[0]  # nmol in a unit cell
        volume = self.cell.volume()  # volume of a unit cell in nm**3
        self.bondlen = None

        try:
            self.bondlen = lat.bondlen
            self.logger.info("Bond length (specified): {0}".format(self.bondlen))
        except AttributeError:
            self.logger.debug("  Estimating the bond threshold length...")
            # assume that the particles distribute homogeneously.
            rc = (volume / nmol)**(1 / 3) * 1.5
            grid = pl.determine_grid(self.cell.mat, rc)
            p = pl.pairs_fine(self.waters, rc, self.cell.mat, grid, distance=False)
            self.bondlen = 1.1 * shortest_distance(self.waters, self.cell, pairs=p)
            self.logger.info("Bond length (estim.): {0}".format(self.bondlen))

        # Set density
        mass = 18  # water
        NB = 6.022e23
        density0 = mass * nmol / (NB * volume * 1e-21)

        if density <= 0:
            try:
                self.density = lat.density
            except AttributeError:
                self.logger.info(
                    "Density is not specified. Assume the density from lattice.")
                dmin = shortest_distance(self.waters, self.cell)
                self.logger.info(
                    "Closest pair distance: {0} (should be around 0.276 nm)".format(dmin))
                self.density = density0 / (0.276 / dmin)**3
                # self.density = density0
        else:
            self.density = density

        self.logger.info("Target Density: {0}".format(self.density))
        self.logger.info("Original Density: {0}".format(density0))

        # scale the cell according to the (specified) density
        ratio = (density0 / self.density)**(1.0 / 3.0)
        self.cell.scale(ratio)

        if self.bondlen is not None:
            self.bondlen *= ratio
        self.logger.info("Bond length (scaled, nm): {0}".format(self.bondlen))

        # ================================================================
        # double_network: True or False
        #   This is a special option for ices VI and VII that have
        #   interpenetrating double network.
        #   GenIce's fast depolarization algorithm fails in some case.
        #
        try:
            self.double_network = lat.double_network
        except AttributeError:
            self.double_network = False

        # ================================================================
        # cages: positions of the centers of cages
        #   In fractional coordinate.
        #
        self.cagepos = None
        self.cagetype = None

        if "cages" in lat.__dict__:
            self.cagepos, self.cagetype = parse_cages(lat.cages)

        # ================================================================
        # fixed: specify the bonds whose directions are fixed.
        #   you can specify them in pairs at a time.
        #   You can also leave it undefined.
        #
        self.fixed = []
        try:
            self.fixed = parse_pairs(lat.fixed)
            self.logger.info("Orientations of some edges are fixed.")
        except AttributeError:
            pass

        if "dopeIonsToUnitCell" in lat.__dict__:
            self.dopeIonsToUnitCell = lat.dopeIonsToUnitCell
        else:
            self.dopeIonsToUnitCell = None
        self.dopants = set()

        # if asis, make pairs to be fixed.
        if self.asis and len(self.fixed) == 0:
            self.fixed = self.pairs

        # filled cages
        self.filled_cages = set()

        # groups info
        self.groups = defaultdict(dict)

        # groups for the semi-guest
        # experimental; there are many variation of semi-guest inclusion.
        self.groups_placer = {"Bu-": butyl,
                              "Butyl-": butyl,
                              "Pentyl-": pentyl,
                              "Propyl-": propyl,
                              "2,2-dimethylpropyl-": _2_2_dimethylpropyl,
                              "2,3-dimethylbutyl-": _2_3_dimethylbutyl,
                              "3,3-dimethylbutyl-": _3_3_dimethylbutyl,
                              "3-methylbutyl-": _3_methylbutyl,
                              "Ethyl-": ethyl}

    def generate_ice(self,
                     water_type,
                     guests,
                     formatter,
                     record_depolarization_path=None,
                     depolarize=True,
                     noise=0.):

        hooks = formatter.hooks
        arg   = formatter.arg
        maxstage = max(0, *hooks.keys())

        if 0 in hooks:
            hooks[0](self, arg)
        elif arg != "":
            logger.info("Arguments are given but the module does not accept them.")
            if "usage" in formatter.__dict__:
                formatter.usage()
            else:
                for line in formatter.__doc__.splitlines():
                    logger.info("  "+line)

        self.stage1(noise)

        if 1 in hooks:
            abort = hooks[1](self)
            if maxstage < 2 or abort:
                return

        res = self.stage2()

        if 2 in hooks:
            abort = hooks[2](self)
            if maxstage < 3 or abort:
                return

        self.stage3()

        if 3 in hooks:
            abort = hooks[3](self)
            if maxstage < 4 or abort:
                return

        self.stage4(depolarize=depolarize,
                    record_depolarization_path=record_depolarization_path)

        if 4 in hooks:
            abort = hooks[4](self)
            if maxstage < 5 or abort:
                return

        self.stage5()

        if 5 in hooks:
            abort = hooks[5](self)
            if maxstage < 6 or abort:
                return

        self.stage6(water_type)

        if 6 in hooks:
            abort = hooks[6](self)
            if maxstage < 7 or abort:
                return

        self.stage7(guests)

        if 7 in hooks:
            hooks[7](self)


    def stage1(self,
               noise=0.):
        """
        Replicate water molecules to make a repeated cell

        Provided variables:
        repposition: replicated molecular positions (CoM, relative)
        repcell:     replicated simulation cell shape matrix
        repcagetype: replicated cage types array
        repcagepos:  replicated cage positions (CoM, relative)
        cagetypes:   set of cage types
        """

        self.logger.info("Stage1: Replication.")
        self.reppositions = replicate_positions(self.waters, self.rep)

        # This must be done before the replication of the cell.
        self.logger.info("  Number of water molecules: {0}".format(
            len(self.reppositions)))
        self.graph = self.prepare_random_graph(self.fixed)

        # scale the cell
        self.repcell = Cell(self.cell.mat)
        self.repcell.scale2(self.rep)

        if noise > 0.0:
            self.logger.info("  Add noise: {0}.".format(noise))
            perturb = np.random.normal(loc=0.0,
                                       scale=noise * 0.01 * 3.0 * 0.5,  # in percent, radius of water
                                       size=self.reppositions.shape)
            self.reppositions += self.repcell.abs2rel(perturb)

        if self.cagepos is not None:
            self.logger.info("  Hints:")
            self.repcagepos = replicate_positions(self.cagepos, self.rep)
            nrepcages = self.repcagepos.shape[0]
            self.repcagetype = [self.cagetype[i % len(self.cagetype)]
                                for i in range(nrepcages)]
            self.cagetypes = defaultdict(set)

            for i, typ in enumerate(self.repcagetype):
                self.cagetypes[typ].add(i)

            # INFO for cage types
            self.logger.info("    Cage types: {0}".format(list(self.cagetypes)))

            for typ, cages in self.cagetypes.items():
                self.logger.info("    Cage type {0}: {1}".format(typ, cages))
            # Up here move to stage 1.

        self.logger.info("Stage1: end.")

    def stage2(self):
        """
        Make a random graph and replicate.

        Provided variables:
        dopants:
        groups:  replicated positions of the chemical groups (CoM, relative)
        filled_cages:
        graph:   replicated network topology (bond orientation may be random)
        """

        self.logger.info("Stage2: Graph preparation.")

        # Some edges are directed when ions are doped.
        if self.dopeIonsToUnitCell is not None:
            self.dopeIonsToUnitCell(self)  # may be defined in the plugin

        # Replicate the dopants in the unit cell
        self.dopants = replicate_labeldict(
            self.dopants, len(self.waters), self.rep)
        self.groups = replicate_groups(
            self.groups, self.waters, self.cagepos, self.rep)

        # self.groups_info(self.groups)
        for root, cages in self.groups.items():
            self.filled_cages |= set(cages)

        # self.logger.info(("filled",self.filled_cages))
        # Replicate the graph
        self.graph = replicate_graph(self.graph, self.waters, self.rep)

        # Dope ions by options.
        if len(self.anions) > 0:
            self.logger.info("  Anionize: {0}.".format(self.anions))

            for site, name in self.anions.items():
                self.graph.anionize(site)
                self.dopants[site] = name

        if len(self.cations) > 0:
            self.logger.info("  Cationize: {0}.".format(self.cations))

            for site, name in self.cations.items():
                self.graph.cationize(site)
                self.dopants[site] = name

        # Count bonds
        nrandom = 0
        nfixed = 0
        for i, j, data in self.graph.edges(data=True):
            if self.graph[i][j]['fixed']:  # fixed pair
                nfixed += 1
            else:
                nrandom += 1
        self.logger.info("  Number of pre-oriented hydrogen bonds: {0}".format(nfixed))
        self.logger.info("  Number of unoriented hydrogen bonds: {0}".format(nrandom))
        self.logger.info("  Number of hydrogen bonds: {0} (regular num: {1})".format(nfixed + nrandom, len(self.reppositions) * 2))

        # test2==True means it is a z=4 graph.
        self.test2 = self.test_undirected_graph(self.graph)
        if not self.test2:
            self.logger.warn("Test2 failed.")

        self.logger.info("Stage2: end.")
        return self.test2

    def stage3(self):
        """
        Make a true ice graph.

        Provided variables:
        graph: network obeying B-F rule.
        """

        self.logger.info("Stage3: Bernal-Fowler rule.")

        if self.asis:
            self.logger.info("  Skip applying the ice rule by request.")
        else:
            self.graph.purge_ice_defects()

        self.logger.info("Stage3: end.")

    def stage4(self, depolarize=True, record_depolarization_path=None):
        """
        Depolarize.

        Provided variables:
        spacegraph: depolarized network with node positions.
        yapresult:  Animation of the depolarization process in YaPlot format.
        """

        self.logger.info("Stage4: Depolarization.")

        if not depolarize or self.asis:
            self.logger.info("  Skip depolarization by request. {0} {1}".format(depolarize, self.asis))
            self.yapresult = ""
            self.spacegraph = dg.SpaceIceGraph(self.graph,
                                               coord=self.reppositions,
                                               ignores=self.graph.ignores)
        else:
            if self.double_network:
                if (self.rep[0] % 2 == 0) and (self.rep[1] % 2 == 0) and (self.rep[2] % 2 == 0):
                    pass
                else:
                    self.logger.error("In making the ice structure having the double network (e.g. ices 6 and 7), all the repetition numbers (--rep) must be even.")
                    sys.exit(1)
            self.spacegraph = dg.SpaceIceGraph(self.graph,
                                               coord=self.reppositions,
                                               ignores=self.graph.ignores)
            if record_depolarization_path is not None:
                draw = dg.YaplotDraw(self.reppositions, self.repcell.mat, data=self.spacegraph)
                yapresult = dg.depolarize(self.spacegraph, self.repcell.mat, draw=draw)
                record_depolarization_path.write(yapresult)
            else:
                dg.depolarize(self.spacegraph, self.repcell.mat, draw=None)
        self.logger.info("Stage4: end.")

    def stage5(self):
        """
        Prepare orientations for rigid molecules.

        Provided variables:
        reppositions: molecular positions.
        rotmatrices:  rotation matrices for water molecules
        """

        self.logger.info("Stage5: Orientation.")

        # determine the orientations of the water molecules based on edge
        # directions.
        self.rotmatrices = orientations(
            self.reppositions, self.spacegraph, self.repcell)

        # Activate it.
        # logger.info("The network is not specified.  Water molecules will be orinented randomly.")
        # rotmatrices = [rigid.rand_rotation_matrix() for pos in positions]
        self.logger.info("Stage5: end.")

    def stage6(self, water_type):
        """
        Arrange water atoms and replacements

        Provided variables:
        atoms: atomic positions of water molecules. (absolute)
        """

        self.logger.info("Stage6: Atomic positions of water.")

        # assert audit_name(water_type), "Dubious water name: {0}".format(water_type)
        # water = importlib.import_module("genice.molecules."+water_type)
        water = safe_import("molecule", water_type)

        try:
            mdoc = water.__doc__.splitlines()
        except BaseException:
            mdoc = []

        for line in mdoc:
            self.logger.info("  "+line)

        self.atoms = arrange_atoms(self.reppositions,
                                   self.repcell,
                                   self.rotmatrices,
                                   water.sites,
                                   water.labels,
                                   water.name,
                                   ignores=set(self.dopants))

        self.logger.info("Stage6: end.")

    def stage7(self, guests):
        """
        Arrange guest atoms

        Provided variables:
        atoms: atomic positions of all molecules.
        """

        self.logger.info("Stage7: Atomic positions of the guest.")

        if self.cagepos is not None:

            # the cages around the dopants.
            dopants_neighbors = self.dopants_info(
                self.dopants, self.reppositions, self.repcagepos, self.repcell)

            # put the (one-off) groups
            if len(self.spot_groups) > 0:
                # process the -H option
                for cage, group_to in self.spot_groups.items():
                    group, root = group_to.split(":")
                    self.add_group(cage, group, int(root))

            molecules = defaultdict(list)

            if len(self.spot_guests) > 0:

                # process the -G option
                for cage, molec in self.spot_guests.items():
                    molecules[molec].append(cage)
                    self.filled_cages.add(cage)

            if guests is not None:

                # process the -g option
                for arg in guests:
                    logging.debug(arg[0])
                    cagetype, spec = arg[0].split("=")
                    assert cagetype in self.cagetypes, "Nonexistent cage type: {0}".format(cagetype)
                    resident = dict()
                    rooms = list(self.cagetypes[cagetype] - self.filled_cages)

                    for room in rooms:
                        resident[room] = None

                    # spec contains a formula consisting of "+" and "*"
                    contents = spec.split("+")
                    vacant = len(rooms)

                    for content in contents:

                        if "*" in content:
                            molec, frac = content.split("*")
                            frac = float(frac)
                        else:
                            molec = content
                            frac = 1.0

                        nmolec = int(frac * len(rooms) + 0.5)
                        vacant -= nmolec
                        assert vacant >= 0, "Too many guests."
                        remain = nmolec
                        movedin = []

                        while remain > 0:
                            r = random.randint(0, len(rooms) - 1)
                            room = rooms[r]

                            if resident[room] is None:
                                resident[room] = molec
                                molecules[molec].append(room)
                                movedin.append(room)
                                remain -= 1

            # Now ge got the address book of the molecules.
            if len(molecules):
                self.logger.info("  Summary of guest placements:")
                self.guests_info(self.cagetypes, molecules)

            if len(self.spot_groups) > 0:
                self.logger.info("  Summary of groups:")
                self.groups_info(self.groups)

            # semi-guests
            for root, cages in self.groups.items():
                assert root in self.dopants
                name = self.dopants[root]
                molname = "G{0}".format(root)
                pos = self.reppositions[root]
                rot = self.rotmatrices[root]
                self.atoms.append([0, molname, name, self.repcell.rel2abs(pos), 0])
                del self.dopants[root]  # processed.
                self.logger.debug((root, cages, name, molname, pos, rot))

                for cage, group in cages.items():
                    assert group in self.groups_placer
                    assert cage in dopants_neighbors[root]
                    cpos = self.repcagepos[cage]
                    self.atoms += self.groups_placer[group](cpos,
                                                            pos,
                                                            self.repcell,
                                                            molname)

            # molecular guests
            for molec, cages in molecules.items():
                gmol = safe_import("molecule", molec)

                try:
                    mdoc = gmol.__doc__.splitlines()
                except BaseException:
                    mdoc = []
                for line in mdoc:
                    logger.info("  "+line)
                cpos = [self.repcagepos[i] for i in cages]
                cmat = [np.identity(3) for i in cages]
                self.atoms += arrange_atoms(cpos, self.repcell,
                                            cmat, gmol.sites, gmol.labels, gmol.name)

        # Assume the dopant is monatomic and replaces one water molecule
        atomset = defaultdict(set)
        for label, name in self.dopants.items():
            atomset[name].add(label)

        for name, labels in atomset.items():
            pos = [self.reppositions[i] for i in sorted(labels)]
            rot = [self.rotmatrices[i] for i in sorted(labels)]
            self.atoms += arrange_atoms(pos,
                                        self.repcell,
                                        rot,
                                        [[0., 0., 0.], ],
                                        [name],
                                        name)

        self.logger.info("Stage7: end.")

    def prepare_random_graph(self, fixed):

        if self.pairs is None:
            self.logger.info("  Pairs are not given explicitly.")
            self.logger.info("  Estimating the bonds according to the pair distances.")

            self.logger.debug("Bondlen: {0}".format(self.bondlen))
            # make bonded pairs according to the pair distance.
            # make before replicating them.
            grid = pl.determine_grid(self.cell.mat, self.bondlen)
            assert np.product(grid) > 0, "Too thin unit cell. Consider use of --rep option if the cell was made by cif2ice."
            self.pairs = pl.pairs_fine(self.waters, self.bondlen, self.cell.mat, grid, distance=False)

            # self.pairs = [v for v in zip(j0,j1)]
            # Check using a simpler algorithm.
            # Do not use it for normal debug because it is too slow
            if False:  # self.logger.level <= logging.DEBUG:
                pairs1 = self.pairs
                pairs2 = [v for v in pl.pairs_crude(self.waters, self.bondlen, self.cell.mat, distance=False)]
                self.logger.debug("pairs1: {0}".format(len(pairs1)))
                self.logger.debug("pairs2: {0}".format(len(pairs2)))
                for pair in pairs1:
                    i, j = pair
                    assert (i, j) in pairs2 or (j, i) in pairs2
                for pair in pairs2:
                    i, j = pair
                    assert (i, j) in pairs1 or (j, i) in pairs1

        graph = dg.IceGraph()
        if fixed is not None:
            for i, j in fixed:
                graph.add_edge(i, j, fixed=True)

        # Fixed pairs are default.
        for pair in self.pairs:
            i, j = pair

            if graph.has_edge(i, j) or graph.has_edge(j, i):
                pass
            else:
                if random.randint(0, 1) == 0:
                    graph.add_edge(i, j, fixed=False)
                else:
                    graph.add_edge(j, i, fixed=False)

        self.logger.info("  Number of water nodes: {0}".format(graph.number_of_nodes()))

        return graph

    def test_undirected_graph(self, graph):
        # Test

        undir = graph.to_undirected()
        for node in range(undir.number_of_nodes()):
            if node not in undir:
                self.logger.debug("z=0 at {0}".format(node))
            else:
                z = len(list(undir.neighbors(node)))
                if z != 4:
                    self.logger.debug("z={0} at {1}".format(z, node))

        if graph.number_of_edges() != len(self.reppositions) * 2:
            self.logger.info("Inconsistent number of HBs {0} for number of molecules {1}.".format(
                graph.number_of_edges(), len(self.reppositions)))
            return False

        return True

    def dopants_info(self, dopants=None, waters=None, cagepos=None, cell=None):
        if dopants is None:
            dopants = self.dopants

        if waters is None:
            waters = self.waters

        if cagepos is None:
            cagepos = self.cagepos

        if cell is None:
            cell = self.cell

        dopants_neighbors = neighbor_cages_of_dopants(dopants, waters, cagepos, cell)

        for dopant, cages in dopants_neighbors.items():
            self.logger.info(
                "    Cages adjacent to dopant {0}: {1}".format(dopant, cages))

        return dopants_neighbors

    def groups_info(self, groups):
        for root, cages in groups.items():
            for cage, group in cages.items():
                self.logger.info(
                    "    Group {0} of dopant {1} in cage {2}".format(group, root, cage))

    def guests_info(self, cagetypes, molecules):
        for cagetype, cageid in cagetypes.items():
            self.logger.info("    Guests in cage type {0}:".format(cagetype))

            for molec, cages in molecules.items():
                cages = set(cages)
                cages &= cageid

                if len(cages):
                    self.logger.info("      {0} * {1} @ {2}".format(molec, len(cages), cages))

    def add_group(self, cage, group, root):
        self.groups[root][cage] = group
        self.filled_cages.add(cage)

    def __del__(self):
        self.logger.info("Completed.")




import logging
import re

import numpy as np
import pairlist as pl

def load_iter(file, oname, hname=None):
    logger = logging.getLogger()
    if __name__[-4:] == "mdva":
        conv = 0.052917721067  # au in nm
    else:
        conv = 0.1             # AA in nm
    while True:
        line = file.readline()  # 1st line:comment
        if len(line) == 0:
            return
        cols = line.split()
        assert cols[0] == '#'  # yaga style
        c = [float(x) for x in cols[1:4]]
        cellmat = np.array([[c[0], 0., 0.],
                            [0., c[1], 0.],
                            [0., 0., c[2]]])
        cellmat *= conv
        celli = np.linalg.inv(cellmat)
        while True:
            line = file.readline()
            if line[0] != "-":
                break
        
        natom = int(line)
        hatoms = []
        oatoms = []
        skipped = set()
        for i in range(natom):
            line = file.readline()
            cols = line.split()
            atomname = cols[0]
            # atomid = int(line[15:20])
            pos = np.array([float(x) for x in cols[1:4]]) * conv
            if re.fullmatch(oname, atomname):
                oatoms.append(pos)
            elif hname is not None and re.fullmatch(hname, atomname):
                hatoms.append(pos)
            else:
                if atomname not in skipped:
                    logger.info("Skip {0}".format(atomname))
                    skipped.add(atomname)
        # fractional coordinate
        oatoms = np.array(oatoms) @ celli
        if len(hatoms) > 0:
            hatoms = np.array(hatoms) @ celli
        else:
            hatoms = None

        yield oatoms, hatoms, cellmat



import logging
import sys
import numpy as np
from genice.cell   import Cell
from genice.genice import GenIce, put_in_array, shortest_distance
from genice.valueparsers import parse_pairs
import pairlist as pl
from collections import defaultdict
from genice import digraph as dg

class AnalIce(GenIce):
    def __init__(self, lat):

        self.logger = logging.getLogger()
        self.rep = (1,1,1)
        density = 0.0
        #         cations=dict(),
        #         anions=dict(),
        #         spot_guests=dict(),
        #         spot_groups=dict(),
        self.asis = False
        # Show the document of the module

        try:
            self.doc = lat.__doc__.splitlines()
        except BaseException:
            self.doc = []

        self.doc.append("")
        self.doc.append("Command line: {0}".format(" ".join(sys.argv)))

        for line in self.doc:
            self.logger.info("  "+line)
        # ================================================================
        # rotmatrices (analice)
        #
        try:
            self.rotmatrices = lat.rotmat
        except BaseException:
            self.logger.info("No rotmatrices in lattice")
            self.rotmatrices = None
        # ================================================================
        # waters: positions of water molecules
        #
        self.waters = put_in_array(lat.waters)
        self.logger.debug("Waters: {0}".format(len(self.waters)))
        self.waters = self.waters.reshape((self.waters.size // 3, 3))

        # ================================================================
        # cell: cell dimension
        #   see parse_cell for syntax.
        #
        self.cell = Cell(lat.cell)

        # ================================================================
        # coord: "relative" or "absolute"
        #   Inside genice, molecular positions are always treated as "relative"
        #
        if lat.coord == "absolute":
            self.waters = self.cell.abs2rel(self.waters)

        self.waters = np.array([w - np.floor(w) for w in self.waters])

        # ================================================================
        # pairs: specify the pairs of molecules that are connected.
        #   Bond orientation will be shuffled later
        #   unless it is "fixed".
        #
        self.pairs = None

        try:
            self.pairs = parse_pairs(lat.pairs)
        except AttributeError:
            self.logger.info("HB connectivity is not defined.")

        # ================================================================
        # bondlen: specify the bond length threshold.
        #   This is used when "pairs" are not specified.
        #   It is applied to the original positions of molecules (before density setting).
        #
        nmol = self.waters.shape[0]  # nmol in a unit cell
        volume = self.cell.volume()  # volume of a unit cell in nm**3
        self.bondlen = None

        try:
            self.bondlen = lat.bondlen
            self.logger.info("Bond length (specified): {0}".format(self.bondlen))
        except AttributeError:
            self.logger.debug("  Estimating the bond threshold length...")
            # assume that the particles distribute homogeneously.
            rc = (volume / nmol)**(1 / 3) * 1.5
            grid = pl.determine_grid(self.cell.mat, rc)
            p = pl.pairs_fine(self.waters, rc, self.cell.mat, grid, distance=False)
            self.bondlen = 1.1 * shortest_distance(self.waters, self.cell, pairs=p)
            self.logger.info("Bond length (estim.): {0}".format(self.bondlen))

        # Set density
        mass = 18  # water
        NB = 6.022e23
        density0 = mass * nmol / (NB * volume * 1e-21)

        if density <= 0:
            try:
                self.density = lat.density
            except AttributeError:
                self.logger.info(
                    "Density is not specified. Assume the density from lattice.")
                dmin = shortest_distance(self.waters, self.cell)
                self.logger.info(
                    "Closest pair distance: {0} (should be around 0.276 nm)".format(dmin))
                self.density = density0 / (0.276 / dmin)**3
                # self.density = density0
        else:
            self.density = density

        self.logger.info("Target Density: {0}".format(self.density))
        self.logger.info("Original Density: {0}".format(density0))

        # scale the cell according to the (specified) density
        ratio = (density0 / self.density)**(1.0 / 3.0)
        self.cell.scale(ratio)

        if self.bondlen is not None:
            self.bondlen *= ratio
        self.logger.info("Bond length (scaled, nm): {0}".format(self.bondlen))

        # ================================================================
        # double_network: True or False
        #   This is a special option for ices VI and VII that have
        #   interpenetrating double network.
        #   GenIce's fast depolarization algorithm fails in some case.
        #
        try:
            self.double_network = lat.double_network
        except AttributeError:
            self.double_network = False

        # ================================================================
        # cages: positions of the centers of cages
        #   In fractional coordinate.
        #
        self.cagepos = None
        self.cagetype = None

        # ================================================================
        # fixed: specify the bonds whose directions are fixed.
        #   you can specify them in pairs at a time.
        #   You can also leave it undefined.
        #
        self.fixed = []
        try:
            self.fixed = parse_pairs(lat.fixed)
            self.logger.info("Orientations of some edges are fixed.")
        except AttributeError:
            pass

        self.dopeIonsToUnitCell = None
        self.dopants = set()

        # if asis, make pairs to be fixed.
        if self.asis and len(self.fixed) == 0:
            self.fixed = self.pairs

        # filled cages
        self.filled_cages = set()

        # groups info
        self.groups = defaultdict(dict)

        
    def analyze_ice(self, water_type, formatter, noise=0.):
        """
        Protocol for analice
        """
        hooks = formatter.hooks
        arg   = formatter.arg

        maxstage = max(0, *hooks.keys())

        if 0 in hooks:
            hooks[0](self, arg)
        elif arg != "":
            logger.info("Arguments are given but the module does not accept them.")
            if "usage" in formatter.__dict__:
                formatter.usage()
            else:
                for line in formatter.__doc__.splitlines():
                    logger.info("  "+line)

        self.stage1(noise)

        if 1 in hooks:
            hooks[1](self)
            if maxstage < 2:
                return

        if self.rotmatrices is None:
            res = self.stage2()

        if 2 in hooks:
            hooks[2](self)
            if maxstage < 3:
                return

        if self.rotmatrices is None:
            self.stage3()

        if 3 in hooks:
            hooks[3](self)
            if maxstage < 4:
                return

        self.stage4()

        if 4 in hooks:
            hooks[4](self)
            if maxstage < 5:
                return

        # molecular orientation should be given in the loader.
        if self.rotmatrices is None:
            self.stage5()

        if 5 in hooks:
            hooks[5](self)
            if maxstage < 6:
                return

        self.stage6(water_type)

        if 6 in hooks:
            hooks[6](self)
            if maxstage < 7:
                return

        # self.stage7_analice(guests)
        if 7 in hooks:
            hooks[7](self)

    def stage1(self,
               noise=0.):
        """
        Do nothing.

        Provided variables:
        repposition: replicated molecular positions (CoM, relative)
        repcell:     replicated simulation cell shape matrix
        """

        self.logger.info("Stage1: (...)")
        self.reppositions = self.waters

        # This must be done before the replication of the cell.
        self.logger.info("  Number of water molecules: {0}".format(len(self.reppositions)))

        # self.graph = self.prepare_random_graph(self.fixed)
        self.graph = self.prepare_random_graph(self.pairs)

        # scale the cell
        self.repcell = Cell(self.cell.mat)

        # self.repcell.scale2(self.rep)
        # add small perturbations to the molecular positions.
        if noise > 0.0:
            self.logger.info("  Add noise: {0}.".format(noise))
            perturb = np.random.normal(loc=0.0,
                                       scale=noise * 0.01 * 3.0 * 0.5,  # in percent, radius of water
                                       size=self.reppositions.shape)
            self.reppositions += self.repcell.abs2rel(perturb)

        self.logger.info("Stage1: end.")

    def stage4(self):
        """
        Depolarize.

        Provided variables:
        spacegraph: depolarized network with node positions.
        yapresult:  Animation of the depolarization process in YaPlot format.
        """

        self.logger.info("Stage4: (...)")
        self.yapresult = ""
        self.spacegraph = dg.SpaceIceGraph(self.graph,
                                           coord=self.reppositions,
                                           ignores=self.graph.ignores)
        self.logger.info("Stage4: end.")



import logging

import numpy as np


def load_iter(file, **kwargs):
    logger = logging.getLogger()
    logger.info("  Loading AR3Rs...")
    cellmat = None
    while True:
        line = file.readline()
        if len(line) == 0:
            return
        if len(line) > 4:
            if line[:5] == "@BOX3":
                logger.info("  @BOX3")
                line = file.readline()
                box = np.array([float(x) for x in line.split()[:3]])
                cellmat = np.diag(box) / 10  # in nm
            elif line[:5] == "@AR3R":
                line = file.readline()
                nmol = int(line.split()[0])
                oatoms = []
                logger.info("  @AR3R")
                for i in range(nmol):
                    line = file.readline()
                    cols = line.split()[:3]
                    pos = np.array([float(x) for x in cols[:3]])
                    oatoms.append(pos)
                hatoms = None
                oatoms = np.array(oatoms)
                yield oatoms, hatoms, cellmat



import re
import importlib
import logging
import pkg_resources as pr
import sys
#
# Accept options parenthesized after the plugin name.
#


def audit_name(name):
    """
    Audit the mol name to avoid the access to external files
    """
    return re.match('^[A-Za-z0-9-_]+$', name) is not None


def import_extra(category, name):
    logger = logging.getLogger()
    logger.info("Extra {0} plugin: {1}".format(category,name))
    groupname = 'genice_{0}'.format(category)
    for ep in pr.iter_entry_points(group=groupname):
        logger.debug("    Entry point: {0}".format(ep))
        if ep.name == name:
            logger.debug("      Loading {0}...".format(name))
            module = ep.load()
    return module



def safe_import(category, name):
    logger = logging.getLogger()
    assert category in ("lattice", "format", "molecule", "loader")
    # name may contain arguments
    left = name.find("[")
    right = name.rfind("]")
    arg = ""
    if left < right:
        arg = name[left + 1:right]
        name = name[:left]
    assert audit_name(name), "Dubious {0} name: {1}".format(category, name)

#    if category == 'format':
#        logger.info("Load {0} module '{1}', arguments [{2}]".format(category, name, arg))
#        return import_format_plugin(category, name), arg

    module = None
    try:
        module = importlib.import_module(category + "s." + name)  # at ~/.genice
    except ImportError as e:
        pass
    if module is None:
        fullname = "genice." + category + "s." + name
        logger.debug("Load module: {0}".format(fullname))
        try:
            module = importlib.import_module(fullname)
        except:
            pass
    if module is None:
        module = import_extra(category, name)
        
    logger.info("Load {0} module '{1}', arguments [{2}]".format(category, name, arg))
    module.logger = logger
    module.arg    = arg

    if category == "format":
        return module
    logger.debug(category)
    # every plugin should have an argparser.
    if "argparser" in module.__dict__:
        # argparser for format plugin should be invoked in different place,
        # after the definition of a lattice object, as a hook.
        module.argparser(arg)
    elif arg != "":
        logger.info("Arguments are given but the module does not accept them.")
        if "usage" in module.__dict__:
            module.usage()
        else:
            for line in module.__doc__.splitlines():
                logger.info("  "+line)
    return module



# coding: utf-8
"""
Yaplot format.
defined in https://github.com/vitroid/Yaplot
"""

from collections import defaultdict
import numpy as np
import yaplotlib as yp


def hook0(lattice, args):
    lattice.logger.info("Hook0: ArgParser.")
    lattice.yaplot_size_H = 0.01
    if args != "":
        for arg in args.split(":"):
            cols = arg.split("=")
            if len(cols) == 2:
                if cols[0] == "H":
                    lattice.yaplot_size_H = float(cols[1])
                    continue
            assert False, "Unknown option."
    lattice.logger.info("Hook0: end.")
    
    



def hook1(lattice):
    lattice.logger.info("Hook1: Draw the cell in Yaplot format.")
    s = yp.Layer(2)
    x,y,z = lattice.repcell.mat
    for p,q,r in ((x,y,z),(y,z,x),(z,x,y)):
        for a in (np.zeros(3), p, q, p+q):
            s += yp.Line(a,a+r)
    print(s,end="")
    lattice.logger.info("Hook1: end.")


def hook2(lattice):
    global nwateratoms
    if lattice.yaplot_size_H > 0:
        return
    lattice.logger.info("Hook2: Output CoM of water molecules in Yaplot format.")
    # prepare the reverse dict
    waters = defaultdict(dict)
    pos = lattice.reppositions @ lattice.repcell.mat
    s = ""
    for p in pos:
        s += yp.Layer(4)
        s += yp.Color(3)
        s += yp.Size(0.03)
        s += yp.Circle(p)
    print(s, end="")
    lattice.logger.info("Hook2: end.")
    return True


def hook6(lattice):
    global nwateratoms
    lattice.logger.info("Hook6: Output water molecules in Yaplot format.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    # prepare the reverse dict
    waters = defaultdict(dict)
    for atom in lattice.atoms:
        resno, resname, atomname, position, order = atom
        if "O" in atomname:
            waters[order]["O"] = position
        elif "H" in atomname:
            if "H0" not in waters[order]:
                waters[order]["H0"] = position
            else:
                waters[order]["H1"] = position
    s = ""
    s += yp.Color(3)
    for order, water in waters.items():
        O = water["O"]        
        H0 = water["H0"]        
        H1 = water["H1"]        
        s += yp.Layer(4)
        s += yp.Color(3)
        s += yp.Size(0.03)
        s += yp.Circle(O)
        s += yp.Line(O,H0)
        s += yp.Line(O,H1)
        s += yp.Size(lattice.yaplot_size_H)
        s += yp.Circle(H0)
        s += yp.Circle(H1)
        s += yp.Color(2)
        s += yp.Layer(1)
        s += yp.Text(O, "{0}".format(order))
    s += yp.Layer(3)
    s += yp.Color(4)
    s += yp.ArrowType(1)
    s += yp.Size(0.03)
    for i,j in lattice.spacegraph.edges(data=False):
        if i in waters and j in waters:  # edge may connect to the dopant
            O = waters[j]["O"]
            H0 = waters[i]["H0"]
            H1 = waters[i]["H1"]
            d0 = H0 - O
            d1 = H1 - O
            rr0 = np.dot(d0,d0)
            rr1 = np.dot(d1,d1)
            if rr0 < rr1 and rr0 < 0.245**2:
                s += yp.Arrow(H0,O)
            if rr1 < rr0 and rr1 < 0.245**2:
                s += yp.Arrow(H1,O)
    print(s, end="")
    nwateratoms = len(lattice.atoms)
    lattice.logger.info("Hook6: end.")


def hook7(lattice):
    global nwateratoms
    lattice.logger.info("Hook7: Output water molecules in Yaplot format.")
    lattice.logger.info("  Total number of atoms: {0}".format(len(lattice.atoms)))
    gatoms = lattice.atoms[nwateratoms:]
    palettes = dict()
    s = ""
    s += yp.Layer(4)
    s += yp.ArrowType(1)
    H = []
    O  = ""
    for atom in gatoms:
        resno, resname, atomname, position, order = atom
        if atomname in palettes:
            pal = palettes[atomname]
        else:
            pal = 4 + len(palettes)
            palettes[atomname] = pal
        s += yp.Color(pal)
        s += yp.Size(0.04)
        s += yp.Circle(position)
    s = '#' + "\n#".join(lattice.doc) + "\n" + s
    print(s)
    lattice.logger.info("Hook7: end.")
    

hooks = {0:hook0, 7:hook7, 6:hook6, 1:hook1, 2:hook2}
